# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import balance.testutil

import numpy as np
import pandas as pd


class TestBalance_weights_stats(
    balance.testutil.BalanceTestCase,
):
    def test__check_weights_are_valid(self):
        from balance.stats_and_plots.weights_stats import _check_weights_are_valid

        w = [np.inf, 1, 2, 1.0, 0]
        self.assertEqual(_check_weights_are_valid(w), None)
        self.assertEqual(_check_weights_are_valid(np.array(w)), None)
        self.assertEqual(_check_weights_are_valid(pd.Series(w)), None)
        self.assertEqual(_check_weights_are_valid(pd.DataFrame(w)), None)
        self.assertEqual(_check_weights_are_valid(pd.DataFrame({"a": w})), None)
        self.assertEqual(
            _check_weights_are_valid(pd.DataFrame({"a": w, "b": [str(x) for x in w]})),
            None,
        )  # checking only the first column

        with self.assertRaisesRegex(TypeError, "weights \\(w\\) must be a number*"):
            _check_weights_are_valid([str(x) for x in w])

        with self.assertRaisesRegex(TypeError, "weights \\(w\\) must be a number*"):
            w = ["a", "b"]
            _check_weights_are_valid(w)

        with self.assertRaisesRegex(
            ValueError, "weights \\(w\\) must all be non-negative values."
        ):
            w = [-1, 0, 1]
            _check_weights_are_valid(w)

    def test_design_effect(self):
        from balance.stats_and_plots.weights_stats import design_effect

        self.assertEqual(design_effect(pd.Series((1, 1, 1, 1))), 1)
        self.assertEqual(
            design_effect(pd.Series((0, 1, 2, 3))),
            1.555_555_555_555_555_6,
        )
        self.assertEqual(type(design_effect(pd.Series((0, 1, 2, 3)))), np.float64)

    def test_nonparametric_skew(self):
        from balance.stats_and_plots.weights_stats import nonparametric_skew

        self.assertEqual(nonparametric_skew(pd.Series((1, 1, 1, 1))), 0)
        self.assertEqual(nonparametric_skew(pd.Series((1))), 0)
        self.assertEqual(nonparametric_skew(pd.Series((1, 2, 3, 4))), 0)
        self.assertEqual(nonparametric_skew(pd.Series((1, 1, 1, 2))), 0.5)

    def test_prop_above_and_below(self):
        from balance.stats_and_plots.weights_stats import prop_above_and_below

        self.assertEqual(
            prop_above_and_below(pd.Series((1, 1, 1, 1))).astype(int).to_list(),
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        )
        self.assertEqual(
            prop_above_and_below(pd.Series((1, 2, 3, 4))).to_list(),
            [0.0, 0.0, 0.0, 0.25, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0],
        )

        self.assertEqual(
            prop_above_and_below(
                pd.Series((1, 2, 3, 4)), below=(0.1, 0.5), above=(2, 3)
            ).to_list(),
            [0.0, 0.25, 0.0, 0.0],
        )
        self.assertEqual(
            prop_above_and_below(
                pd.Series((1, 2, 3, 4)), below=(0.1, 0.5), above=(2, 3)
            ).index.to_list(),
            ["prop(w < 0.1)", "prop(w < 0.5)", "prop(w >= 2)", "prop(w >= 3)"],
        )

        self.assertEqual(
            prop_above_and_below(pd.Series((1, 2, 3, 4)), above=None, below=None), None
        )

        # test return_as_series = False
        self.assertEqual(
            {
                k: v.to_list()
                for k, v in prop_above_and_below(
                    pd.Series((1, 2, 3, 4)), return_as_series=False
                ).items()
            },
            {"below": [0.0, 0.0, 0.0, 0.25, 0.5], "above": [0.5, 0.0, 0.0, 0.0, 0.0]},
        )

    def test_weighted_median_breakdown_point(self):
        import pandas as pd
        from balance.stats_and_plots.weights_stats import (
            weighted_median_breakdown_point,
        )

        self.assertEqual(weighted_median_breakdown_point(pd.Series((1, 1, 1, 1))), 0.5)
        self.assertEqual(weighted_median_breakdown_point(pd.Series((2, 2, 2, 2))), 0.5)
        self.assertEqual(
            weighted_median_breakdown_point(pd.Series((1, 1, 1, 10))), 0.25
        )
        self.assertEqual(
            weighted_median_breakdown_point(pd.Series((1, 1, 1, 1, 10))), 0.2
        )


class TestBalance_weighted_stats(
    balance.testutil.BalanceTestCase,
):
    def test__prepare_weighted_stat_args(self):
        from balance.stats_and_plots.weighted_stats import _prepare_weighted_stat_args

        v, w = [-1, 0, 1, np.inf], [np.inf, 1, 2, 1.0]
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)

        v, w = pd.Series([-1, 0, 1, np.inf]), pd.Series([np.inf, 1, 2, 1.0])
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)
        # check the values are the same:
        self.assertEqual(v.to_list(), v2[0].to_list())
        self.assertEqual(w.to_list(), w2.to_list())
        # Do we have any inf or nan?
        self.assertTrue(any(np.isinf(v2[0])))
        self.assertTrue(any(np.isinf(w2)))
        self.assertTrue(not any(np.isnan(v2[0])))
        self.assertTrue(not any(np.isnan(w2)))

        # Checking how it works when inf_rm = True
        v2, w2 = _prepare_weighted_stat_args(v, w, True)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)
        # check the values are the NOT same (because the Inf was turned to nan):
        self.assertNotEqual(v.to_list(), v2[0].to_list())
        self.assertNotEqual(w.to_list(), w2.to_list())
        # Do we have any inf or nan?
        self.assertTrue(not any(np.isinf(v2[0])))
        self.assertTrue(not any(np.isinf(w2)))
        self.assertTrue(any(np.isnan(v2[0])))
        self.assertTrue(any(np.isnan(w2)))

        # Check that it catches wrong input types
        with self.assertRaises(TypeError):
            v, w = pd.Series([1, 2]), np.matrix([1, 2]).T
            v2, w2 = _prepare_weighted_stat_args(v, w)
        with self.assertRaises(TypeError):
            v, w = pd.Series([1, 2]), (1, 2)
            v2, w2 = _prepare_weighted_stat_args(v, w)
        with self.assertRaises(TypeError):
            v, w = (1, 2), pd.Series([1, 2])
            v2, w2 = _prepare_weighted_stat_args(v, w)
        with self.assertRaises(ValueError):
            v, w = pd.Series([1, 2, 3]), pd.Series([1, 2])
            v2, w2 = _prepare_weighted_stat_args(v, w)
        with self.assertRaises(TypeError):
            v, w = pd.Series(["a", "b"]), pd.Series([1, 2])
            v2, w2 = _prepare_weighted_stat_args(v, w)

        # check other input types for v and w
        # np.array
        v, w = np.array([-1, 0, 1, np.inf]), np.array([np.inf, 1, 2, 1.0])
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)
        # pd.DataFrame
        v, w = pd.DataFrame([-1, 0, 1, np.inf]), np.array([np.inf, 1, 2, 1.0])
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)
        # np.matrix
        v, w = np.matrix([-1, 0, 1, np.inf]).T, np.array([np.inf, 1, 2, 1.0])
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        # assert the new types:
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)

        # Dealing with w=None
        v, w = pd.Series([-1, 0, 1, np.inf]), None
        v2, w2 = _prepare_weighted_stat_args(v, w, False)
        self.assertEqual(type(v2), pd.DataFrame)
        self.assertEqual(type(w2), pd.Series)
        self.assertEqual(w2, pd.Series([1.0, 1.0, 1.0, 1.0]))
        self.assertTrue(len(w2) == 4)
        # Verify defaults:
        self.assertEqual(
            _prepare_weighted_stat_args(v, None, False)[0],
            _prepare_weighted_stat_args(v)[0],
        )

        with self.assertRaises(ValueError):
            v, w = pd.Series([-1, 0, 1, np.inf]), pd.Series([np.inf, 1, -2, 1.0])
            v2, w2 = _prepare_weighted_stat_args(v, w)

    def test_weighted_mean(self):
        from balance.stats_and_plots.weighted_stats import weighted_mean

        # No weights
        self.assertEqual(weighted_mean(pd.Series([-1, 0, 1, 2])), pd.Series(0.5))
        self.assertEqual(weighted_mean(pd.Series([-1, None, 1, 2])), pd.Series(0.5))

        # No weights, with one inf value
        self.assertEqual(
            weighted_mean(pd.Series([-1, np.inf, 1, 2])),
            pd.Series(np.inf),
        )
        self.assertEqual(
            weighted_mean(pd.Series([-1, np.inf, 1, 2]), inf_rm=True),
            pd.Series(0.5),
        )

        # Inf value in weights
        self.assertTrue(
            all(
                np.isnan(
                    weighted_mean(
                        pd.Series([-1, 2, 1, 2]), w=pd.Series([1, np.inf, 1, 1])
                    )
                )
            )
        )
        self.assertEqual(
            weighted_mean(
                pd.Series([-1, 2, 1, 2]), w=pd.Series([1, np.inf, 1, 1]), inf_rm=True
            ),
            pd.Series(2 / 3),
        )

        # With weights
        self.assertEqual(
            weighted_mean(pd.Series([-1, 2, 1, 2]), w=pd.Series([1, 2, 3, 4])),
            pd.Series(1.4),
        )

        self.assertEqual(
            weighted_mean(pd.Series([-1, 2, 1, 2]), w=pd.Series([1, None, 1, 1])),
            pd.Series(2 / 3),
        )

        # Notice that while None values are ignored, their weights will still be used in the denominator
        # Hence, None in values acts like 0
        self.assertEqual(
            weighted_mean(pd.Series([-1, None, 1, 2]), w=pd.Series([1, 0, 1, 1])),
            pd.Series(2 / 3),
        )
        self.assertEqual(
            weighted_mean(pd.Series([-1, np.nan, 1, 2]), w=pd.Series([1, 0, 1, 1])),
            pd.Series(2 / 3),
        )
        self.assertEqual(
            weighted_mean(pd.Series([-1, None, 1, 2]), w=pd.Series([1, 1, 1, 1])),
            pd.Series(1 / 2),
        )
        # None in the values is just like 0
        self.assertEqual(
            weighted_mean(pd.Series([-1, None, 1, 2]), w=pd.Series([1, 1, 1, 1])),
            weighted_mean(pd.Series([-1, 0, 1, 2]), w=pd.Series([1, 1, 1, 1])),
        )

        # pd.DataFrame v
        d = pd.DataFrame([(-1, 2, 1, 2), (1, 2, 3, 4)]).transpose()
        pd.testing.assert_series_equal(weighted_mean(d), pd.Series((1, 2.5)))
        pd.testing.assert_series_equal(
            weighted_mean(d, w=pd.Series((1, 2, 3, 4))),
            pd.Series((1.4, 3)),
        )

        # np.matrix v
        d = np.matrix([(-1, 2, 1, 2), (1, 2, 3, 4)]).transpose()
        pd.testing.assert_series_equal(weighted_mean(d), pd.Series((1, 2.5)))
        pd.testing.assert_series_equal(
            weighted_mean(d, w=pd.Series((1, 2, 3, 4))),
            pd.Series((1.4, 3)),
        )

        #  Test input validation
        with self.assertRaisesRegex(TypeError, "must be numeric"):
            weighted_mean(pd.Series(["a", "b"]))
        with self.assertRaisesRegex(TypeError, "must be numeric"):
            weighted_mean(pd.DataFrame({"a": [1, 2], "b": ["a", "b"]}))

        with self.assertRaisesRegex(ValueError, "must have same number of rows"):
            weighted_mean(pd.Series([1, 2]), pd.Series([1, 2, 3]))
        with self.assertRaisesRegex(ValueError, "must have same number of rows"):
            weighted_mean(pd.DataFrame({"a": [1, 2]}), pd.Series([1]))

        # Make sure that index is ignored
        self.assertEqual(
            weighted_mean(
                pd.Series([-1, 2, 1, 2], index=(0, 1, 2, 3)),
                w=pd.Series([1, 2, 3, 4], index=(5, 6, 7, 8)),
                inf_rm=True,
            ),
            pd.Series((-1 * 1 + 2 * 2 + 1 * 3 + 2 * 4) / (1 + 2 + 3 + 4)),
        )

        self.assertEqual(
            weighted_mean(
                pd.Series([-1, 2, 1, 2], index=(0, 1, 2, 3)),
                w=pd.Series([1, 2, 3, 4], index=(0, 6, 7, 8)),
                inf_rm=True,
            ),
            pd.Series((-1 * 1 + 2 * 2 + 1 * 3 + 2 * 4) / (1 + 2 + 3 + 4)),
        )

    def var_of_weighted_mean(self):
        from balance.stats_and_plots.weighted_stats import var_of_weighted_mean

        # Test no weights assigned
        #  In R: sum((1:4 - mean(1:4))^2 / 4) / (4)
        #  [1] 0.3125
        self.assertEqual(
            var_of_weighted_mean(pd.Series((1, 2, 3, 4))), pd.Series(0.3125)
        )

        #  For a reproducible R example, see: https://gist.github.com/talgalili/b92cd8cdcbfc287e331a8f27db265c00
        self.assertEqual(
            var_of_weighted_mean(pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4))),
            pd.Series(0.24),
        )

    def ci_of_weighted_mean(self):
        from balance.stats_and_plots.weighted_stats import ci_of_weighted_mean

        self.assertEqual(
            ci_of_weighted_mean(pd.Series((1, 2, 3, 4)), round_ndigits=3).to_list(),
            [(1.404, 3.596)],
        )

        self.assertEqual(
            ci_of_weighted_mean(
                pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4)), round_ndigits=3
            ).to_list(),
            [(2.04, 3.96)],
        )

        df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [1, 1, 1, 1]})
        w = pd.Series((1, 2, 3, 4))
        self.assertEqual(
            ci_of_weighted_mean(df, w, conf_level=0.99, round_ndigits=3).to_dict(),
            {"a": (1.738, 4.262), "b": (1.0, 1.0)},
        )

    def test_weighted_var(self):
        from balance.stats_and_plots.weighted_stats import weighted_var

        #  > var(c(1, 2, 3, 4))
        #  [1] 1.66667
        self.assertEqual(weighted_var(pd.Series((1, 2, 3, 4))), pd.Series(5 / 3))

        #  > SDMTools::wt.var(c(1, 2), c(1, 2))
        #  [1] 0.5
        self.assertEqual(
            weighted_var(pd.Series((1, 2)), pd.Series((1, 2))), pd.Series(0.5)
        )

    def test_weighted_sd(self):
        from balance.stats_and_plots.weighted_stats import weighted_sd

        #  > sd(c(1, 2, 3, 4))
        #  [1] 1.290994
        self.assertEqual(weighted_sd(pd.Series((1, 2, 3, 4))), pd.Series(1.290_994))

        #  > SDMTools::wt.sd(c(1, 2), c(1, 2))
        #  [1] 0.7071068
        self.assertEqual(
            weighted_sd(pd.Series((1, 2)), pd.Series((1, 2))),
            pd.Series(0.707_106_8),
        )

        x = [1, 2, 3, 4]
        x2 = pd.Series(x)
        manual_std = np.sqrt(np.sum((x2 - x2.mean()) ** 2) / (len(x) - 1))
        self.assertEqual(round(weighted_sd(x)[0], 5), round(manual_std, 5))

    def test_weighted_quantile(self):
        from balance.stats_and_plots.weighted_stats import weighted_quantile

        self.assertEqual(
            weighted_quantile(np.arange(1, 100, 1), 0.5).values,
            np.array(((50,),)),
        )

        # In R: reldist::wtd.quantile(c(1, 2, 3), q=c(0.5, 0.75), weight=c(1, 1, 2))
        self.assertEqual(
            weighted_quantile(np.array([1, 2, 3]), (0.5, 0.75)).values,
            np.array(((2,), (3,))),
        )

        self.assertEqual(
            weighted_quantile(np.array([1, 2, 3]), 0.5, np.array([1, 1, 2])).values,
            np.percentile([1, 2, 3, 3], 50),
        )

        # verify it indeed works with pd.DataFrame input
        # no weights
        self.assertEqual(
            weighted_quantile(
                pd.DataFrame([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
            ).values,
            np.array([[1.5, 1.0], [2.5, 1.0]]),
        )
        # with weights
        self.assertEqual(
            weighted_quantile(
                pd.DataFrame([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
                w=pd.Series([1, 1, 0, 0]),
            ).values,
            np.array([[1.0, 1.0], [1.5, 1.0]]),
        )
        self.assertEqual(
            weighted_quantile(
                pd.DataFrame([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
                w=pd.Series([1, 100, 1, 1]),
            ).values,
            np.array([[2.0, 1.0], [2.0, 1.0]]),
        )

        # verify it indeed works with np.matrix input
        # no weights
        self.assertEqual(
            weighted_quantile(
                np.matrix([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
            ).values,
            np.array([[1.5, 1.0], [2.5, 1.0]]),
        )
        # with weights
        self.assertEqual(
            weighted_quantile(
                np.matrix([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
                w=pd.Series([1, 1, 0, 0]),
            ).values,
            np.array([[1.0, 1.0], [1.5, 1.0]]),
        )
        self.assertEqual(
            weighted_quantile(
                np.matrix([[1, 2, 3, 4], [1, 1, 1, 1]]).transpose(),
                [0.25, 0.5],
                w=pd.Series([1, 100, 1, 1]),
            ).values,
            np.array([[2.0, 1.0], [2.0, 1.0]]),
        )

    def test_descriptive_stats(self):
        from balance.stats_and_plots.weighted_stats import (
            descriptive_stats,
            weighted_mean,
            weighted_sd,
        )
        from statsmodels.stats.weightstats import DescrStatsW

        x = pd.Series([-1, 0, 1, 2])
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), stat="mean"),
            weighted_mean(x).to_frame().T,
        )
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), stat="std"), weighted_sd(x).to_frame().T
        )

        x = [1, 2, 3, 4]
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), stat="var_of_mean").to_dict(),
            {0: {0: 0.3125}},
        )

        # with weights
        x, w = pd.Series([-1, 2, 1, 2]), pd.Series([1, 2, 3, 4])
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), w, stat="mean"),
            weighted_mean(x, w).to_frame().T,
        )
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), w, stat="std"),
            weighted_sd(x, w).to_frame().T,
        )
        # show that with/without weights gives different results
        self.assertNotEqual(
            descriptive_stats(pd.DataFrame(x), stat="mean").iloc[0, 0],
            descriptive_stats(pd.DataFrame(x), w, stat="mean").iloc[0, 0],
        )
        self.assertNotEqual(
            descriptive_stats(pd.DataFrame(x), stat="std").iloc[0, 0],
            descriptive_stats(pd.DataFrame(x), w, stat="std").iloc[0, 0],
        )
        # shows that descriptive_stats can calculate std_mean and that it's smaller than std (as expected.)
        self.assertTrue(
            descriptive_stats(pd.DataFrame(x), w, stat="std").iloc[0, 0]
            > descriptive_stats(pd.DataFrame(x), w, stat="std_mean").iloc[0, 0]
        )

        x = [1, 2, 3, 4]
        x2 = pd.Series(x)
        tmp_sd = np.sqrt(np.sum((x2 - x2.mean()) ** 2) / (len(x) - 1))
        tmp_se = tmp_sd / np.sqrt(len(x))
        self.assertEqual(
            round(descriptive_stats(pd.DataFrame(x), stat="std_mean").iloc[0, 0], 5),
            round(tmp_se, 5),
        )

        # verify DescrStatsW can deal with None weights
        self.assertEqual(
            DescrStatsW([1, 2, 3], weights=[1, 1, 1]).mean,
            DescrStatsW([1, 2, 3], weights=None).mean,
        )
        self.assertEqual(
            DescrStatsW([1, 2, 3], weights=None).mean,
            2.0,
        )

        x, w = [1, 2, 3, 4], [1, 2, 3, 4]
        self.assertEqual(
            descriptive_stats(pd.DataFrame(x), w, stat="var_of_mean").to_dict(),
            {0: {0: 0.24}},
        )
        self.assertEqual(
            descriptive_stats(
                pd.DataFrame(x), w, stat="ci_of_mean", conf_level=0.99, round_ndigits=3
            ).to_dict(),
            {0: {0: (1.738, 4.262)}},
        )


class TestBalance_weighted_comparisons_stats(
    balance.testutil.BalanceTestCase,
):
    def test_outcome_variance_ratio(self):
        from balance.stats_and_plots.weighted_comparisons_stats import (
            outcome_variance_ratio,
        )

        np.random.seed(876324)

        d = pd.DataFrame(np.random.rand(1000, 11))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghijk"[i] for i in range(0, 11)})

        self.assertEqual(
            outcome_variance_ratio(d[["j", "k"]], d[["j", "k"]]),
            pd.Series([1.0, 1.0], index=["j", "k"]),
        )

    def test__weights_per_covars_names(self):
        # toy example
        from balance.stats_and_plots.weighted_comparisons_stats import (
            _weights_per_covars_names,
        )

        asmd_df = pd.DataFrame(
            {
                "age": 0.5,
                "education[T.high_school]": 1,
                "education[T. bachelor]": 1,
                "education[T. masters]": 1,
                "education[T. phd]": 1,
            },
            index=("self",),
        )

        outcome = _weights_per_covars_names(asmd_df.columns.values.tolist()).to_dict()
        # We expect a df with 2 columns: weight and main_covar_names.
        expected = {
            "weight": {
                "age": 1.0,
                "education[T.high_school]": 0.25,
                "education[T. bachelor]": 0.25,
                "education[T. masters]": 0.25,
                "education[T. phd]": 0.25,
            },
            "main_covar_names": {
                "age": "age",
                "education[T.high_school]": "education",
                "education[T. bachelor]": "education",
                "education[T. masters]": "education",
                "education[T. phd]": "education",
            },
        }

        self.assertEqual(outcome, expected)

    def test_asmd(self):
        from balance.stats_and_plots.weighted_comparisons_stats import asmd

        with self.assertRaisesRegex(ValueError, "sample_df must be pd.DataFrame, is*"):
            # Using wild card since it will return:
            # "sample_df must be pd.DataFrame, is* <class 'pandas.core.series.Series'>"
            asmd(
                pd.Series((0, 1, 2, 3)),
                pd.Series((0, 1, 2, 3)),
                pd.Series((0, 1, 2, 3)),
                pd.Series((0, 1, 2, 3)),
            )

        with self.assertRaisesRegex(ValueError, "target_df must be pd.DataFrame, is*"):
            asmd(
                pd.DataFrame({"a": (0, 1, 2, 3)}),
                pd.Series((0, 1, 2, 3)),
                pd.Series((0, 1, 2, 3)),
                pd.Series((0, 1, 2, 3)),
            )

        # If column names are different, it will only calculate asmd for
        # the overlapping columns. The rest will be np.nan.
        # The mean(asmd) will be calculated while treating the nan values as 0s.
        self.assertEqual(
            np.isnan(
                asmd(
                    pd.DataFrame({"a": (1, 2), "b": (-1, 12)}),
                    pd.DataFrame({"a": (3, 4), "c": (5, 6)}),
                )
            ).tolist(),
            [False, True, True, False],
        )

        with self.assertRaisesRegex(ValueError, "std_type must be in*"):
            asmd(
                pd.DataFrame({"a": (1, 2), "b": (-1, 12)}),
                pd.DataFrame({"a": (3, 4), "c": (5, 6)}),
                std_type="magic variance type that doesn't exist",
            )

        # TODO: (p2) add comparison to the following numbers
        # Benchmark for the numbers:
        # ---------------------------
        #  Test data from R cobalt package:
        #  cobalt::bal.tab(
        #    data.frame(a=c(1, 2, 3, 4), b=c(-1, 12, 0, 42)),
        #    treat=c(1, 1, 0, 0),
        #    weights=c(1, 2, 1, 2),
        #    s.d.denom='control',
        #    method='weighting'
        #  )
        # Output:
        # Balance Measures
        #      Type Diff.Adj
        # a Contin.  -2.8284
        # b Contin.  -0.6847

        # Effective sample sizes
        #            Control Treated
        # Unadjusted     2.      2.
        # Adjusted       1.8     1.8

        # show the default "target" calculation is working as expected
        a1 = pd.Series((1, 2))
        b1 = pd.Series((-1, 1))
        a2 = pd.Series((3, 4))
        b2 = pd.Series((-2, 2))
        w1 = pd.Series((1, 1))
        w2 = w1
        r = asmd(
            pd.DataFrame({"a": a1, "b": b1}),
            pd.DataFrame({"a": a2, "b": b2}),
            w1,
            w2,
        ).to_dict()
        exp_a = np.abs(a1.mean() - a2.mean()) / a2.std()
        exp_b = np.abs(b1.mean() - b2.mean()) / b2.std()
        self.assertEqual((r["a"], r["b"]), (exp_a, exp_b))
        # show that the default is weights equal 1.
        r_no_weights = asmd(
            pd.DataFrame({"a": a1, "b": b1}),
            pd.DataFrame({"a": a2, "b": b2}),
        ).to_dict()
        self.assertEqual(r, r_no_weights)
        # demonstrate that weights effect the outcome (use 0 weights.)
        a1 = pd.Series((1, 2, 100))
        b1 = pd.Series((-1, 1, 100))
        a2 = pd.Series((3, 4, 100))
        b2 = pd.Series((-2, 2, 100))
        w1 = pd.Series((1, 1, 0))
        w2 = w1
        r_with_0_3rd_weight = asmd(
            pd.DataFrame({"a": a1, "b": b1}),
            pd.DataFrame({"a": a2, "b": b2}),
            w1,
            w2,
        ).to_dict()
        self.assertEqual(r, r_with_0_3rd_weight)

        # Not passing weights is the same as weights of all 1s
        r = asmd(
            pd.DataFrame({"a": (1, 2), "b": (-1, 12)}),
            pd.DataFrame({"a": (3, 4), "b": (0, 42)}),
            pd.Series((1, 2)),
            pd.Series((1, 2)),
        )
        e_asmd = pd.Series(
            (2.828_427_1, 0.684_658_9, (2.828_427_1 + 0.684_658_9) / 2),
            index=("a", "b", "mean(asmd)"),
        )
        self.assertEqual(r, e_asmd)
        self.assertEqual(type(r), pd.Series)

        r = asmd(
            pd.DataFrame({"a": (1, 2), "b": (-1, 12)}),
            pd.DataFrame({"a": (3, 4), "b": (0, 42)}),
            pd.Series((1, 2)),
            pd.Series((1, 2)),
        )
        e_asmd = pd.Series(
            (2.828_427_1, 0.684_658_9, (2.828_427_1 + 0.684_658_9) / 2),
            index=("a", "b", "mean(asmd)"),
        )
        self.assertEqual(r, e_asmd)

        # Test different std_types
        args = (
            pd.DataFrame({"a": (1, 2), "b": (-1, 12)}),
            pd.DataFrame({"a": (3, 4), "b": (0, 42)}),
            pd.Series((1, 2)),
            pd.Series((1, 2)),
        )

        r = asmd(*args, std_type="target")
        self.assertEqual(r, e_asmd)

        # TODO: this should be called : test consistency of asmd
        r = asmd(*args, std_type="sample")
        e_asmd_sample = pd.Series(
            (
                2.828_427_124_746_189_4,
                2.211_975_059_096_379,
                (2.828_427_124_746_189_4 + 2.211_975_059_096_379) / 2,
            ),
            index=("a", "b", "mean(asmd)"),
        )
        self.assertEqual(r, e_asmd_sample)

        r = asmd(*args, std_type="pooled")
        e_asmd_pooled = pd.Series(
            (2.828_427_1, 0.924_959_4, (2.828_427_1 + 0.924_959_4) / 2),
            index=("a", "b", "mean(asmd)"),
        )
        self.assertEqual(r, e_asmd_pooled)

        #  Test with categoricals
        #  Categorical variable has missing value in one df
        r = asmd(
            pd.DataFrame({"c": ("x", "y", "x", "x"), "a": (5, 6, 7, 8)}),
            pd.DataFrame({"c": ("x", "y", "x", "z"), "a": (1, 2, 3, 4)}),
            pd.Series((1, 2, 3, 4)),
            pd.Series((1, 1, 1, 1)),
        )
        e_asmd = pd.Series(
            (
                3.485_685_011_586_675_3,
                0.519_615_242_270_663_2,
                0.099_999_999_999_999_98,
                np.nan,
                (
                    3.485_685_011_586_675_3
                    + 0.519_615_242_270_663_2 * (1 / 3)
                    + 0.099_999_999_999_999_98 * (1 / 3)
                    + 0 * (1 / 3)
                )
                / (2),
            ),
            index=("a", "c[x]", "c[y]", "c[z]", "mean(asmd)"),
        )
        self.assertEqual(r, e_asmd)

        # Check that using aggregate_by_main_covar works
        a1 = pd.Series((1, 2))
        b1_A = pd.Series((1, 3))
        b1_B = pd.Series((-1, -3))
        a2 = pd.Series((3, 4))
        b2_A = pd.Series((2, 3))
        b2_B = pd.Series((-2, -3))
        w1 = pd.Series((1, 1))
        w2 = w1

        r1 = asmd(
            pd.DataFrame({"a": a1, "b[A]": b1_A, "b[B]": b1_B}),
            pd.DataFrame({"a": a2, "b[A]": b2_A, "b[B]": b2_B}),
            w1,
            w2,
        ).to_list()

        r2 = asmd(
            pd.DataFrame({"a": a1, "b[A]": b1_A, "b[B]": b1_B}),
            pd.DataFrame({"a": a2, "b[A]": b2_A, "b[B]": b2_B}),
            w1,
            w2,
            "target",
            True,
        ).to_list()

        self.assertTrue(
            all((np.round(r1, 5)) == np.array([2.82843, 0.70711, 0.70711, 1.76777]))
        )
        self.assertTrue(all((np.round(r2, 5)) == np.array([2.82843, 0.70711, 1.76777])))

    def test__aggregate_asmd_by_main_covar(self):
        from balance.stats_and_plots.weighted_comparisons_stats import (
            _aggregate_asmd_by_main_covar,
        )

        # toy example
        asmd_series = pd.Series(
            {
                "age": 0.5,
                "education[T.high_school]": 1,
                "education[T. bachelor]": 2,
                "education[T. masters]": 3,
                "education[T. phd]": 4,
            }
        )

        outcome = _aggregate_asmd_by_main_covar(asmd_series).to_dict()
        expected = {"age": 0.5, "education": 2.5}

        self.assertEqual(outcome, expected)

    def test_asmd_improvement(self):
        from balance.stats_and_plots.weighted_comparisons_stats import asmd_improvement

        r = asmd_improvement(
            pd.DataFrame({"a": (5, 6, 7, 8)}),
            pd.DataFrame({"a": (5, 6, 7, 8)}),
            pd.DataFrame({"a": (1, 2, 3, 4)}),
            pd.Series((1, 1, 1, 1)),
            pd.Series((2, 2, 2, 2)),
            pd.Series((1, 1, 1, 1)),
        )
        self.assertEqual(r, 0)
        self.assertEqual(type(r), np.float64)

        r = asmd_improvement(
            pd.DataFrame({"a": (5, 6, 7, 8)}),
            pd.DataFrame({"a": (1, 2, 3, 4)}),
            pd.DataFrame({"a": (1, 2, 3, 4)}),
            pd.Series((1, 1, 1, 1)),
            pd.Series((1, 1, 1, 1)),
            pd.Series((1, 1, 1, 1)),
        )
        self.assertEqual(r, 1)
        self.assertEqual(type(r), np.float64)

        r = asmd_improvement(
            pd.DataFrame({"a": (3, 4)}),
            pd.DataFrame({"a": (2, 3)}),
            pd.DataFrame({"a": (1, 2)}),
            pd.Series((1, 1)),
            pd.Series((1, 1)),
            pd.Series((1, 1)),
        )
        self.assertEqual(r, 0.5)
        self.assertEqual(type(r), np.float64)


class TestBalance_general_stats(
    balance.testutil.BalanceTestCase,
):
    def test_relative_response_rates(self):
        from balance.stats_and_plots.general_stats import relative_response_rates

        df = pd.DataFrame(
            {"o1": (7, 8, 9, 10), "o2": (7, 8, 9, np.nan), "id": (1, 2, 3, 4)}
        )

        self.assertEqual(
            relative_response_rates(df).to_dict(),
            {
                "o1": {"n": 4.0, "%": 100.0},
                "o2": {"n": 3.0, "%": 75.0},
                "id": {"n": 4.0, "%": 100.0},
            },
        )

        df_target = pd.concat([df, df])
        self.assertEqual(
            relative_response_rates(df, df_target).to_dict(),
            {
                "o1": {"n": 4.0, "%": 50.0},
                "o2": {"n": 3.0, "%": 50.0},
                "id": {"n": 4.0, "%": 50.0},
            },
        )

        # verify behavior when per_column is set to False
        self.assertEqual(
            relative_response_rates(df, df_target, per_column=False).round(3).to_dict(),
            {
                "o1": {"n": 4.0, "%": 66.667},
                "o2": {"n": 3.0, "%": 50.0},
                "id": {"n": 4.0, "%": 66.667},
            },
        )

        with self.assertRaisesRegex(
            ValueError, "df and df_target must have the exact same columns*"
        ):
            relative_response_rates(df, df_target.iloc[:, 0:1], per_column=True)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Dict, List, Optional, Union

import pandas as pd

from balance import adjustment as balance_adjustment, util as balance_util

logger: logging.Logger = logging.getLogger(__package__)


# TODO: Add tests for all arguments of function
# TODO: Add argument for na_action
def poststratify(
    sample_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_df: pd.DataFrame,
    target_weights: pd.Series,
    variables: Optional[List[str]] = None,
    transformations: str = "default",
    transformations_drop: bool = True,
    *args,
    **kwargs,
) -> Dict[str, Union[pd.Series, Dict[str, str]]]:
    """Perform cell-based post-stratification. The output weights take into account
    the design weights and the post-stratification weights.
    Reference: https://docs.wfp.org/api/documents/WFP-0000121326/download/

    Args:
        sample_df (pd.DataFrame): a dataframe representing the sample
        sample_weights (pd.Series): design weights for sample
        target_df (pd.DataFrame): a dataframe representing the target
        target_weights (pd.Series): design weights for target
        variables (Optional[List[str]], optional): list of variables to include in the model.
            If None all joint variables of sample_df and target_df are used
        transformations (str, optional): what transformations to apply to data before fitting the model.
            Default is "default" (see apply_transformations function)
        transformations_drop (bool, optional): whether the function should drop non-transformed variables.
            Default is True.
    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        Dict[str, Union[pd.Series, Dict[str, str]]]:
            weight (pd.Series): final weights (sum up to target's sum of weights)
            model (dict): method of adjustment

            Dict shape:
            {
                "weight": w,
                "model": {"method": "poststratify"},
            }
    """
    balance_util._check_weighting_methods_input(sample_df, sample_weights, "sample")
    balance_util._check_weighting_methods_input(target_df, target_weights, "target")

    if ("weight" in sample_df.columns.values) or ("weight" in target_df.columns.values):
        raise ValueError(
            "weight can't be a name of a column in sample or target when applying poststratify"
        )

    variables = balance_util.choose_variables(sample_df, target_df, variables=variables)
    logger.debug(f"Join variables for sample and target: {variables}")

    sample_df = sample_df.loc[:, variables]
    target_df = target_df.loc[:, variables]

    sample_df, target_df = balance_adjustment.apply_transformations(
        (sample_df, target_df),
        transformations=transformations,
        drop=transformations_drop,
    )
    variables = list(sample_df.columns)
    logger.debug(f"Final variables in the model after transformations: {variables}")

    target_df = target_df.assign(weight=target_weights)
    target_cell_props = target_df.groupby(list(variables))["weight"].sum()

    sample_df = sample_df.assign(design_weight=sample_weights)
    sample_cell_props = sample_df.groupby(list(variables))["design_weight"].sum()

    combined = pd.merge(
        target_cell_props,
        sample_cell_props,
        right_index=True,
        left_index=True,
        how="outer",
    )

    # check that all combinations of cells in sample_df are also in target_df
    if any(combined["weight"].isna()):
        raise ValueError("all combinations of cells in sample_df must be in target_df")

    combined["weight"] = combined["weight"] / combined["design_weight"]
    sample_df = sample_df.join(combined["weight"], on=variables)
    w = sample_df.weight * sample_df.design_weight

    return {
        "weight": w,
        "model": {"method": "poststratify"},
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.


import logging
from typing import Optional

from balance.balancedf_class import (  # noqa
    BalanceCovarsDF,  # noqa
    BalanceOutcomesDF,  # noqa
    BalanceWeightsDF,  # noqa
)
from balance.datasets import load_data  # noqa
from balance.sample_class import Sample  # noqa
from balance.util import TruncationFormatter  # noqa

# TODO: which objects do we want to explicitly externalize?
# TODO: verify this works.

global __version__
__version__ = "0.8.0"


def setup_logging(
    logger_name: Optional[str] = __package__,
    level: str = "INFO",
    removeHandler: bool = True,
) -> logging.Logger:
    """
    Initiates a nicely formatted logger called "balance", with level "info".
    """
    if removeHandler:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    logger = logging.getLogger(logger_name)

    logger.setLevel(getattr(logging, level))
    formatter = TruncationFormatter(
        "%(levelname)s (%(asctime)s) [%(module)s/%(funcName)s (line %(lineno)d)]: %(message)s"
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


logger: logging.Logger = setup_logging()
logger.info(f"Using {__package__} version {__version__}")


# TODO: add example in the notebooks for using this function.
def set_warnings(level: str = "WARNING") -> None:
    logger.setLevel(getattr(logging, level))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import math
from fractions import Fraction

from functools import reduce
from typing import Callable, Dict, List, Union

import numpy as np
import pandas as pd

from balance import adjustment as balance_adjustment, util as balance_util
from ipfn import ipfn

logger = logging.getLogger(__package__)


# TODO: Add options for only marginal distributions input
def rake(
    sample_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_df: pd.DataFrame,
    target_weights: pd.Series,
    variables: Union[List[str], None] = None,
    transformations: Union[Dict[str, Callable], str] = "default",
    na_action: str = "add_indicator",
    max_iteration: int = 1000,
    convergence_rate: float = 0.0005,
    rate_tolerance: float = 1e-8,
    *args,
    **kwargs,
) -> Dict:
    """
    Perform raking (using the iterative proportional fitting algorithm).
    See: https://en.wikipedia.org/wiki/Iterative_proportional_fitting

    Returns weights normalised to sum of target weights

    Arguments:
    sample_df --- (pandas dataframe) a dataframe representing the sample.
    sample_weights --- (pandas series) design weights for sample.
    target_df ---  (pandas dataframe) a dataframe representing the target.
    target_weights --- (pandas series) design weights for target.
    variables ---  (list of strings) list of variables to include in the model.
                   If None all joint variables of sample_df and target_df are used.
    transformations --- (dict) what transformations to apply to data before fitting the model.
                               Default is "default" (see apply_transformations function).
    na_action --- (string) what to do with NAs. Default is "add_indicator", which adds NaN as a
                  group (called "__NaN__") for each weighting variable (post-transformation);
                  "drop" removes rows with any missing values on any variable from both sample
                  and target.
    max_iteration --- (int) maximum number of iterations for iterative proportional fitting algorithm
    convergence_rate --- (float) convergence criteria; the maximum difference in proportions between
                          sample and target marginal distribution on any covariate in order
                          for algorithm to converge.
    rate_tolerance --- (float) convergence criteria; if convergence rate does not move more
                               than this amount than the algorithm is also considered to
                               have converged.

    Returns:
    A dictionary including:
    "weight" --- The weights for the sample.
    "model" --- parameters of the model: iterations (dataframe with iteration numbers and
                convergence rate information at all steps), converged (Flag with the output
                status: 0 for failure and 1 for success).
    """
    assert (
        "weight" not in sample_df.columns.values
    ), "weight shouldn't be a name for covariate in the sample data"
    assert (
        "weight" not in target_df.columns.values
    ), "weight shouldn't be a name for covariate in the target data"

    # TODO: move the input checks into separate funnction for rake, ipw, poststratify
    assert isinstance(sample_df, pd.DataFrame), "sample_df must be a pandas DataFrame"
    assert isinstance(target_df, pd.DataFrame), "target_df must be a pandas DataFrame"
    assert isinstance(
        sample_weights, pd.Series
    ), "sample_weights must be a pandas Series"
    assert isinstance(
        target_weights, pd.Series
    ), "target_weights must be a pandas Series"
    assert sample_df.shape[0] == sample_weights.shape[0], (
        "sample_weights must be the same length as sample_df"
        f"{sample_df.shape[0]}, {sample_weights.shape[0]}"
    )
    assert target_df.shape[0] == target_weights.shape[0], (
        "target_weights must be the same length as target_df"
        f"{target_df.shape[0]}, {target_weights.shape[0]}"
    )

    variables = balance_util.choose_variables(sample_df, target_df, variables=variables)

    logger.debug(f"Join variables for sample and target: {variables}")

    sample_df = sample_df.loc[:, variables]
    target_df = target_df.loc[:, variables]

    assert len(variables) > 1, (
        "Must weight on at least two variables for raking. "
        f"Currently have variables={variables} only"
    )

    sample_df, target_df = balance_adjustment.apply_transformations(
        (sample_df, target_df), transformations
    )

    # TODO: separate into a function that handles NA (for rake, ipw, poststratify)
    if na_action == "drop":
        (sample_df, sample_weights) = balance_util.drop_na_rows(
            sample_df, sample_weights, "sample"
        )
        (target_df, target_weights) = balance_util.drop_na_rows(
            target_df, target_weights, "target"
        )
    elif na_action == "add_indicator":
        target_df = target_df.fillna("__NaN__")
        sample_df = sample_df.fillna("__NaN__")
    else:
        raise ValueError("`na_action` must be 'add_indicator' or 'drop'")

    # Cast all data types as string to be explicit about each unique value
    # being its own group and to handle that `fillna()` above creates
    # series of type Object, which won't work for the ipfn script
    levels_dict = {}
    for variable in variables:
        target_df[variable] = target_df[variable].astype(str)
        sample_df[variable] = sample_df[variable].astype(str)

        sample_var_set = set(sample_df[variable].unique())
        target_var_set = set(target_df[variable].unique())
        sample_over_set = sample_var_set - target_var_set
        target_over_set = target_var_set - sample_var_set
        if len(sample_over_set):
            raise ValueError(
                "All variable levels in sample must be present in target. "
                f"'{variable}' in target is missing these levels: {sample_over_set}."
            )
        if len(target_over_set):
            logger.warning(
                f"'{variable}' has more levels in target than in sample. "
                f"'{variable}' in sample is missing these levels: {target_over_set}. "
                "These levels are treated as if they do not exist for that variable."
            )
        levels_dict[variable] = list(sample_var_set.intersection(target_var_set))

    logger.info(
        f"Final covariates and levels that will be used in raking: {levels_dict}."
    )

    target_df = target_df.assign(weight=target_weights)
    sample_df = sample_df.assign(weight=sample_weights)

    sample_sum_weights = sample_df["weight"].sum()
    target_sum_weights = target_df["weight"].sum()

    # Alphabetize variables to ensure consistency across covariate order
    # (ipfn algorithm is iterative and variable order can matter on the margins)
    alphabetized_variables = list(variables)
    alphabetized_variables.sort()

    logger.debug(
        f"Alphabetized variable order is as follows: {alphabetized_variables}."
    )

    # margins from population
    target_margins = [
        (
            (
                target_df.groupby(variable)["weight"].sum()
                / (sum(target_df.groupby(variable)["weight"].sum()))
                * sample_sum_weights
            )
        )
        for variable in alphabetized_variables
    ]

    # sample cells (joint distribution of covariates)
    sample_cells = (
        sample_df.groupby(alphabetized_variables)["weight"].sum().reset_index()
    )

    logger.debug(
        "Raking algorithm running following settings: "
        f" convergence_rate: {convergence_rate}; max_iteration: {max_iteration}; rate_tolerance: {rate_tolerance}"
    )
    # returns dataframe with joint distribution of covariates and total weight
    # for that specific set of covariates
    ipfn_obj = ipfn.ipfn(
        original=sample_cells,
        aggregates=target_margins,
        dimensions=[[var] for var in alphabetized_variables],
        weight_col="weight",
        convergence_rate=convergence_rate,
        max_iteration=max_iteration,
        verbose=2,
        rate_tolerance=rate_tolerance,
    )
    fit, converged, iterations = ipfn_obj.iteration()

    logger.debug(
        f"Raking algorithm terminated with following convergence: {converged}; "
        f"and iteration meta data: {iterations}."
    )

    if not converged:
        logger.warning("Maximum iterations reached, convergence was not achieved")

    raked = pd.merge(
        sample_df.reset_index(),
        fit.rename(columns={"weight": "rake_weight"}),
        how="left",
        on=list(variables),
    )

    # Merge back to original sample weights
    raked_rescaled = pd.merge(
        raked,
        (
            sample_df.groupby(list(variables))["weight"]
            .sum()
            .reset_index()
            .rename(columns={"weight": "total_survey_weight"})
        ),
        how="left",
        on=list(variables),
    ).set_index(
        "index"
    )  # important if dropping rows due to NA

    # use above merge to give each individual sample its proportion of the
    # cell's total weight
    raked_rescaled["rake_weight"] = (
        raked_rescaled["rake_weight"] / raked_rescaled["total_survey_weight"]
    )
    # rescale weights to sum to target_sum_weights (sum of initial target weights)
    w = (
        raked_rescaled["rake_weight"] / raked_rescaled["rake_weight"].sum()
    ) * target_sum_weights
    return {
        "weight": w,
        "model": {
            "method": "rake",
            "iterations": iterations,
            "converged": converged,
            "perf": {"prop_dev_explained": np.array([np.nan])},
            # TODO: fix functions that use the perf and remove it from here
        },
    }


def _lcm(a: int, b: int) -> int:
    """
    Calculates the least common multiple (LCM) of two integers.

    The least common multiple (LCM) of two or more numbers is the smallest positive integer that is divisible by each of the given numbers.
    In other words, it is the smallest multiple that the numbers have in common.
    The LCM is useful when you need to find a common denominator for fractions or synchronize repeating events with different intervals.

    For example, let's find the LCM of 4 and 6:

    The multiples of 4 are: 4, 8, 12, 16, 20, ...
    The multiples of 6 are: 6, 12, 18, 24, 30, ...
    The smallest multiple that both numbers have in common is 12, so the LCM of 4 and 6 is 12.

    The calculation is based on the property that the product of two numbers is equal to the product of their LCM and GCD: a * b = LCM(a, b) * GCD(a, b).
    (proof: https://math.stackexchange.com/a/589299/1406)

    Args:
        a: The first integer.
        b: The second integer.

    Returns:
        The least common multiple of the two integers.
    """
    # NOTE: this function uses math.gcd which calculates the greatest common divisor (GCD) of two integers.
    # The greatest common divisor (GCD) of two or more integers is the largest positive integer that divides each of the given integers without leaving a remainder.
    # In other words, it is the largest common factor of the given integers.
    # For example, the GCD of 12 and 18 is 6, since 6 is the largest integer that divides both 12 and 18 without leaving a remainder.
    # Similarly, the GCD of 24, 36, and 48 is 12, since 12 is the largest integer that divides all three of these numbers without leaving a remainder.
    return abs(a * b) // math.gcd(a, b)


def _proportional_array_from_dict(
    input_dict: Dict[str, float], max_length: int = 10000
) -> List[str]:
    """
    Generates a proportional array based on the input dictionary.

    Args:
        input_dict (Dict[str, float]): A dictionary where keys are strings and values are their proportions (float).
        max_length (int): check if the length of the output exceeds the max_length. If it does, it will be scaled down to that length. Default is 10k.

    Returns:
        A list of strings where each key is repeated according to its proportion.

    Examples:
        ::
            _proportional_array_from_dict({"a":0.2, "b":0.8})
                # ['a', 'b', 'b', 'b', 'b']
            _proportional_array_from_dict({"a":0.5, "b":0.5})
                # ['a', 'b']
            _proportional_array_from_dict({"a":1/3, "b":1/3, "c": 1/3})
                # ['a', 'b', 'c']
            _proportional_array_from_dict({"a": 3/8, "b": 5/8})
                # ['a', 'a', 'a', 'b', 'b', 'b', 'b', 'b']
            _proportional_array_from_dict({"a":3/5, "b":1/5, "c": 2/10})
                # ['a', 'a', 'a', 'b', 'c']
    """
    # Filter out items with zero value
    filtered_dict = {k: v for k, v in input_dict.items() if v > 0}

    # Normalize the values if they don't sum to 1
    sum_values = sum(filtered_dict.values())
    if sum_values != 1:
        filtered_dict = {k: v / sum_values for k, v in filtered_dict.items()}

    # Convert the proportions to fractions
    fraction_dict = {
        k: Fraction(v).limit_denominator() for k, v in filtered_dict.items()
    }

    # Find the least common denominator (LCD) of the fractions
    lcd = 1
    for fraction in fraction_dict.values():
        lcd *= fraction.denominator // math.gcd(lcd, fraction.denominator)

    # Calculate the scaling factor based on max_length
    scaling_factor = min(1, max_length / lcd)

    result = []
    for key, fraction in fraction_dict.items():
        # Calculate the count for each key based on its proportion and scaling_factor
        k_count = round((fraction * lcd).numerator * scaling_factor)
        # Extend the result array with the key repeated according to its count
        result.extend([key] * k_count)

    return result


def _find_lcm_of_array_lengths(arrays: Dict[str, List[str]]) -> int:
    """
    Finds the least common multiple (LCM) of the lengths of arrays in the input dictionary.

    Args:
        arrays: A dictionary where keys are strings and values are lists of strings.

    Returns:
        The LCM of the lengths of the arrays in the input dictionary.

    Example:
        ::
            arrays = {
                        "v1": ["a", "b", "b", "c"],
                        "v2": ["aa", "bb"]
                    }
            _find_lcm_of_array_lengths(arrays)
                # 4

            arrays = {
                        "v1": ["a", "b", "b", "c"],
                        "v2": ["aa", "bb"],
                        "v3": ["a1", "a2", "a3"]
                    }
            _find_lcm_of_array_lengths(arrays)
                # 12
    """
    array_lengths = [len(arr) for arr in arrays.values()]
    lcm_length = reduce(lambda x, y: _lcm(x, y), array_lengths)
    return lcm_length


def _realize_dicts_of_proportions(
    dict_of_dicts: Dict[str, Dict[str, float]]
) -> Dict[str, List[str]]:
    """
    Generates proportional arrays of equal length for each input dictionary.

    This can be used to get an input dict of proportions of values, and produce a dict with arrays that realizes these proportions.
    It can be used as input to the Sample object so it could be used for running raking.

    Args:
        dict_of_dicts: A dictionary of dictionaries, where each key is a string and
                   each value is a dictionary with keys as strings and values as their proportions (float).

    Returns:
        A dictionary with the same keys as the input and equal length arrays as values.

    Examples:
        ::
            dict_of_dicts = {
                "v1": {"a": 0.2, "b": 0.6, "c": 0.2},
                "v2": {"aa": 0.5, "bb": 0.5}
            }

            realize_dicts_of_proportions(dict_of_dicts)
            # {'v1': ['a', 'b', 'b', 'b', 'c', 'a', 'b', 'b', 'b', 'c'], 'v2': ['aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb']}


            dict_of_dicts = {
                "v1": {"a": 0.2, "b": 0.6, "c": 0.2},
                "v2": {"aa": 0.5, "bb": 0.5},
                "v3": {"A": 0.2, "B": 0.8},
            }
            realize_dicts_of_proportions(dict_of_dicts)
                # {'v1': ['a', 'b', 'b', 'b', 'c', 'a', 'b', 'b', 'b', 'c'],
                #  'v2': ['aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb'],
                #  'v3': ['A', 'B', 'B', 'B', 'B', 'A', 'B', 'B', 'B', 'B']}
            # The above example could have been made shorter. But that's a limitation of the function.

            dict_of_dicts = {
                "v1": {"a": 0.2, "b": 0.6, "c": 0.2},
                "v2": {"aa": 0.5, "bb": 0.5},
                "v3": {"A": 0.2, "B": 0.8},
                "v4": {"A": 0.1, "B": 0.9},
            }
            realize_dicts_of_proportions(dict_of_dicts)
                # {'v1': ['a', 'b', 'b', 'b', 'c', 'a', 'b', 'b', 'b', 'c'],
                #  'v2': ['aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb', 'aa', 'bb'],
                #  'v3': ['A', 'B', 'B', 'B', 'B', 'A', 'B', 'B', 'B', 'B'],
                #  'v4': ['A', 'B', 'B', 'B', 'B', 'B', 'B', 'B', 'B', 'B']}
    """
    # Generate proportional arrays for each dictionary
    arrays = {k: _proportional_array_from_dict(v) for k, v in dict_of_dicts.items()}

    # Find the LCM over the lengths of all the arrays
    lcm_length = _find_lcm_of_array_lengths(arrays)

    # Extend each array to have the same LCM length while maintaining proportions
    result = {}
    for k, arr in arrays.items():
        factor = lcm_length // len(arr)
        extended_arr = arr * factor
        result[k] = extended_arr

    return result


def prepare_marginal_dist_for_raking(
    dict_of_dicts: Dict[str, Dict[str, float]]
) -> pd.DataFrame:
    """
    Realizes a nested dictionary of proportions into a DataFrame.

    Args:
        dict_of_dicts: A nested dictionary where the outer keys are column names
                   and the inner dictionaries have keys as category labels
                   and values as their proportions (float).

    Returns:
        A DataFrame with columns specified by the outer keys of the input dictionary
        and rows containing the category labels according to their proportions.
        An additional "id" column is added with integer values as row identifiers.

    Example:
        ::
            print(prepare_marginal_dist_for_raking({
                            "A": {"a": 0.5, "b": 0.5},
                            "B": {"x": 0.2, "y": 0.8}
                        }))
            # Returns a DataFrame with columns A, B, and id
                #    A  B  id
                # 0  a  x   0
                # 1  b  y   1
                # 2  a  y   2
                # 3  b  y   3
                # 4  a  y   4
                # 5  b  x   5
                # 6  a  y   6
                # 7  b  y   7
                # 8  a  y   8
                # 9  b  y   9
    """
    target_dict_from_marginals = _realize_dicts_of_proportions(dict_of_dicts)
    target_df_from_marginals = pd.DataFrame.from_dict(target_dict_from_marginals)
    # Add an id column:
    target_df_from_marginals["id"] = range(target_df_from_marginals.shape[0])

    return target_df_from_marginals



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from contextlib import contextmanager
from typing import Any, Callable, cast, Dict, Generator, List, Optional, Tuple, Union

import glmnet_python  # noqa  # Required so that cvglmnet import works

import numpy as np
import pandas as pd
import scipy

from balance import adjustment as balance_adjustment, util as balance_util
from balance.stats_and_plots.weighted_comparisons_stats import asmd
from balance.stats_and_plots.weights_stats import design_effect

from cvglmnet import cvglmnet  # pyre-ignore[21]: this module exists
from cvglmnetPredict import cvglmnetPredict  # pyre-ignore[21]: this module exists
from scipy.sparse.csc import csc_matrix

logger: logging.Logger = logging.getLogger(__package__)


# Allows us to control exactly where monkey patching is applied (e.g.: for better code readability and exceptions tracking).
@contextmanager
def _patch_nan_in_amin_amax(*args, **kwds) -> Generator:
    """This allows us to use nanmin and nanmax instead of amin and amax (thus removing nan from their computation)

    This is needed in cases that the cvglmnet yields some nan values for the cross validation folds.

    Returns:
        Generator: replaces amin and amax, and once done, turns them back to their original version.
    """

    # swap amin and amax with nanmin and nanmax
    # so that they won't return nan when their input has some nan values
    tmp_amin, tmp_amax = scipy.amin, scipy.amax  # pyre-ignore[16]

    # Wrapping amin and amax with logger calls so to alert the user in case nan were present.
    # This comes with the strong assumption that this will occur within the cross validation step!
    def new_amin(a, *args, **kwds) -> Callable:
        nan_count = scipy.count_nonzero(scipy.isnan(a))  # pyre-ignore[16]
        if nan_count > 0:
            logger.warning(
                "The scipy.amin function was replaced with scipy.nanmin."
                f"While running, it removed {nan_count} `nan` values from an array of size {a.size}"
                f"(~{round(100 * nan_count / a.size, 1)}% of the values)."
            )
        return scipy.nanmin(a, *args, **kwds)  # pyre-ignore[16]

    def new_amax(a, *args, **kwds) -> Callable:
        nan_count = scipy.count_nonzero(scipy.isnan(a))  # pyre-ignore[16]
        if nan_count > 0:
            logger.warning(
                "The scipy.amax function was replaced with scipy.nanmax. "
                f"While running, it removed {nan_count} `nan` values from an array of size {a.size}"
                f"(~{round(100 * nan_count / a.size, 1)}% of the values)."
            )
        return scipy.nanmax(a, *args, **kwds)  # pyre-ignore[16]

    scipy.amin = new_amin  # scipy.nanmin
    scipy.amax = new_amax  # scipy.nanmax
    try:
        yield
    finally:
        # undo the function swap
        scipy.amin, scipy.amax = tmp_amin, tmp_amax


@contextmanager
def _patch_scipy_random(*args, **kwds) -> Generator:
    """Monkey-patch scipy.random(), used by glmnet_python
    but removed in scipy 1.9.0"""

    tmp_scipy_random_func = (
        # pyre-ignore[16]
        scipy.random
        if hasattr(scipy, "random")
        else None
    )
    scipy.random = np.random
    try:
        yield
    finally:
        # undo the function swap
        scipy.random = tmp_scipy_random_func


def cv_glmnet_performance(
    fit, feature_names: Optional[list] = None, s: Union[str, float, None] = "lambda_1se"
) -> Dict[str, Any]:
    """Extract elements from cvglmnet to describe the fitness quality.

    Args:
        fit (_type_): output of cvglmnet
        feature_names (Optional[list], optional): The coeficieents of which features should be included.
            None = all features are included. Defaults to None.
        s (Union[str, float, None], optional): lambda avlue for cvglmnet. Defaults to "lambda_1se".

    Raises:
        Exception: _description_

    Returns:
        Dict[str, Any]: Dict of the shape:
            {
                "prop_dev_explained": fit["glmnet_fit"]["dev"][optimal_lambda_index],
                "mean_cv_error": fit["cvm"][optimal_lambda_index],
                "coefs": coefs,
            }
    """
    if isinstance(s, str):
        optimal_lambda = fit[s]
    else:
        optimal_lambda = s
    optimal_lambda_index = np.where(fit["lambdau"] == optimal_lambda)[0]
    if len(optimal_lambda_index) != 1:
        raise Exception(
            f"No lambda found for s={s}. You must specify a "
            'numeric value which exists in the model or "lambda_min", or '
            '"lambda_1se"'
        )
    coefs = cvglmnetPredict(
        fit,
        newx=np.empty([0]),
        ptype="coefficients",
        s=optimal_lambda,
    )
    coefs = coefs.reshape(coefs.shape[0])
    if feature_names is not None:
        coefs = pd.Series(data=coefs, index=["intercept"] + feature_names)
    else:
        coefs = pd.Series(data=coefs)

    return {
        "prop_dev_explained": fit["glmnet_fit"]["dev"][optimal_lambda_index],
        "mean_cv_error": fit["cvm"][optimal_lambda_index],
        "coefs": coefs,
    }


# TODO: consider add option to normalize weights to sample size
def weights_from_link(
    link: Any,
    balance_classes: bool,
    sample_weights: pd.Series,
    target_weights: pd.Series,
    weight_trimming_mean_ratio: Union[None, float, int] = None,
    weight_trimming_percentile: Optional[float] = None,
    keep_sum_of_weights: bool = True,
) -> pd.Series:
    """Transform output of cvglmnetPredict(..., type='link') into weights, by
    exponentiating them, and optionally balancing the classes and trimming
    the weights, then normalize the weights to have sum equal to the sum of the target weights.

    Args:
        link (Any): output of cvglmnetPredict(..., type='link')
        balance_classes (bool): whether balance_classes used in glmnet
        sample_weights (pd.Series): vector of sample weights
        target_weights (pd.Series): vector of sample weights
        weight_trimming_mean_ratio (Union[None, float, int], optional): to be used in :func:`trim_weights`. Defaults to None.
        weight_trimming_percentile (Optional[float], optional): to be used in :func:`trim_weights`. Defaults to None.
        keep_sum_of_weights (bool, optional): to be used in :func:`trim_weights`. Defaults to True.

    Returns:
        pd.Series: A vecotr of normalized weights (for sum of target weights)
    """
    link = link.reshape((link.shape[0],))
    if balance_classes:
        odds = np.sum(sample_weights) / np.sum(target_weights)
        link = link + np.log(odds)
    weights = sample_weights / np.exp(link)
    weights = balance_adjustment.trim_weights(
        weights,
        weight_trimming_mean_ratio,
        weight_trimming_percentile,
        keep_sum_of_weights=keep_sum_of_weights,
    )
    # Normalize weights such that the sum will be the sum of the weights of target
    weights = weights * np.sum(target_weights) / np.sum(weights)
    return weights


# TODO: Update choose_regularization function to be based on mse (instead of grid search)
def choose_regularization(
    fit,
    sample_df: pd.DataFrame,
    target_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_weights: pd.Series,
    X_matrix_sample,
    balance_classes: bool,
    max_de: float = 1.5,
    trim_options: Tuple[
        int, int, int, float, float, float, float, float, float, float
    ] = (20, 10, 5, 2.5, 1.25, 0.5, 0.25, 0.125, 0.05, 0.01),
    n_asmd_candidates: int = 10,
) -> Dict[str, Any]:
    """Searches through the regularisation parameters of the model and weight
    trimming levels to find the combination with the highest covariate
    ASMD reduction (in sample_df and target_df, NOT in the model matrix used for modeling
    the response) subject to the design effect being lower than max_de (deafults to 1.5).
    The function preforms a grid search over the n_asmd_candidates (deafults to 10) models
    with highest DE lower than max_de (assuming higher DE means more bias reduction).

    Args:
        fit (_type_): output of cvglmnet
        sample_df (pd.DataFrame): a dataframe representing the sample
        target_df (pd.DataFrame): a dataframe representing the target
        sample_weights (pd.Series): design weights for sample
        target_weights (pd.Series): design weights for target
        X_matrix_sample (_type_): the matrix that was used to consturct the model
        balance_classes (bool): whether balance_classes used in glmnet
        max_de (float, optional): upper bound for the design effect of the computed weights.
            Used for choosing the model regularization and trimming.
            If set to None, then it uses 'lambda_1se'. Defaults to 1.5.
        trim_options (Tuple[ int, int, int, float, float, float, float, float, float, float ], optional):
            options for weight_trimming_mean_ratio. Defaults to (20, 10, 5, 2.5, 1.25, 0.5, 0.25, 0.125, 0.05, 0.01).
        n_asmd_candidates (int, optional): number of candidates for grid search.. Defaults to 10.

    Returns:
        Dict[str, Any]: Dict of the value of the chosen lambda, the value of trimming, model description.
            Shape is
                {
                    "best": {"s": best.s.values, "trim": best.trim.values[0]},
                    "perf": all_perf,
                }
    """

    logger.info("Starting choosing regularisation parameters")
    # get all links
    links = cvglmnetPredict(fit, X_matrix_sample, ptype="link", s=fit["lambdau"])

    asmd_before = asmd(
        sample_df=sample_df,
        target_df=target_df,
        sample_weights=sample_weights,
        target_weights=target_weights,
    )
    # Grid search over regularisation parameter and weight trimming
    # First calculates design effects for all combinations, because this is cheap
    all_perf = []
    for wr in trim_options:
        for i in range(links.shape[1]):

            s = fit["lambdau"][i]
            link = links[:, i]
            weights = weights_from_link(
                link,
                balance_classes,
                sample_weights,
                target_weights,
                weight_trimming_mean_ratio=wr,
            )

            deff = design_effect(weights)
            all_perf.append(
                {
                    "s": s,
                    "s_index": i,
                    "trim": wr,
                    "design_effect": deff,
                }
            )
    all_perf = pd.DataFrame(all_perf)
    best = (
        # pyre-fixme[16]: `Optional` has no attribute `tail`.
        all_perf[all_perf.design_effect < max_de]
        # pyre-fixme[6]: For 1st param expected `Union[typing_extensions.Literal[0],
        #  typing_extensions.Literal['index']]` but got
        #  `typing_extensions.Literal['design_effect']`.
        .sort_values("design_effect").tail(n_asmd_candidates)
    )
    logger.debug(f"Regularisation with design effect below {max_de}: \n {best}")

    # Calculate ASMDS for best 10 candidates (assuming that higher DE means
    #  more bias reduction)
    all_perf = []
    for _, r in best.iterrows():
        wr = r.trim
        s_index = int(r.s_index)
        s = fit["lambdau"][s_index]
        link = links[:, s_index]

        weights = weights_from_link(
            link,
            balance_classes,
            sample_weights,
            target_weights,
            weight_trimming_mean_ratio=wr,
        )
        adjusted_df = sample_df[sample_df.index.isin(weights.index)]

        asmd_after = asmd(
            # pyre-fixme[6]: For 1st param expected `DataFrame` but got `Series`.
            sample_df=adjusted_df,
            target_df=target_df,
            sample_weights=weights,
            target_weights=target_weights,
        )
        # TODO: use asmd_improvement function for that
        asmd_improvement = (
            asmd_before.loc["mean(asmd)"] - asmd_after.loc["mean(asmd)"]
        ) / asmd_before.loc["mean(asmd)"]
        deff = design_effect(weights)
        all_perf.append(
            {
                "s": s,
                # pyre-fixme[61]: `i` is undefined, or not always defined.
                "s_index": i,
                "trim": wr,
                "design_effect": deff,
                "asmd_improvement": asmd_improvement,
                "asmd": asmd_after.loc["mean(asmd)"],
            }
        )

    all_perf = pd.DataFrame(all_perf)
    best = (
        all_perf[all_perf.design_effect < max_de]
        # pyre-fixme[6]: For 1st param expected `Union[typing_extensions.Literal[0],
        #  typing_extensions.Literal['index']]` but got
        #  `typing_extensions.Literal['asmd_improvement']`.
        .sort_values("asmd_improvement").tail(1)
    )
    logger.info(f"Best regularisation: \n {best}")
    solution = {
        "best": {"s": best.s.values, "trim": best.trim.values[0]},
        "perf": all_perf,
    }
    return solution


# TODO: add memoaization (maybe in the adjust stage?!)
def ipw(
    sample_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_df: pd.DataFrame,
    target_weights: pd.Series,
    variables: Optional[List[str]] = None,
    model: str = "glmnet",
    weight_trimming_mean_ratio: Optional[Union[int, float]] = 20,
    weight_trimming_percentile: Optional[float] = None,
    balance_classes: bool = True,
    transformations: str = "default",
    na_action: str = "add_indicator",
    max_de: Optional[float] = None,
    formula: Union[str, List[str], None] = None,
    penalty_factor: Optional[List[float]] = None,
    one_hot_encoding: bool = False,
    # TODO: This is set to be false in order to keep reproducibility of works that uses balance.
    # The best practice is for this to be true.
    random_seed: int = 2020,
    *args,
    **kwargs,
) -> Dict[str, Any]:
    """Fit an ipw (inverse propensity score weighting) for the sample using the target.

    Args:
        sample_df (pd.DataFrame): a dataframe representing the sample
        sample_weights (pd.Series): design weights for sample
        target_df (pd.DataFrame): a dataframe representing the target
        target_weights (pd.Series): design weights for target
        variables (Optional[List[str]], optional): list of variables to include in the model.
            If None all joint variables of sample_df and target_df are used. Defaults to None.
        model (str, optional): the model used for modeling the propensity scores.
            "glmnet" is logistic model. Defaults to "glmnet".
        weight_trimming_mean_ratio (Optional[Union[int, float]], optional): indicating the ratio from above according to which
            the weights are trimmed by mean(weights) * ratio.
            Defaults to 20.
        weight_trimming_percentile (Optional[float], optional): if weight_trimming_percentile is not none, winsorization is applied.
            if None then trimming is applied. Defaults to None.
        balance_classes (bool, optional): whether to balance the sample and target size for running the model.
            True is preferable for imbalanced cases.
            It is done to make the computation of the glmnet more efficient.
            It shouldn't have an effect on the final weights as this is factored
            into the computation of the weights. TODO: add ref. Defaults to True.
        transformations (str, optional): what transformations to apply to data before fitting the model.
            See apply_transformations function. Defaults to "default".
        na_action (str, optional): what to do with NAs.
            See add_na_indicator function. Defaults to "add_indicator".
        max_de (Optional[float], optional): upper bound for the design effect of the computed weights.
            Used for choosing the model regularization and trimming.
            If set to None, then it uses 'lambda_1se'. Defaults to 1.5.
        formula (Union[str, List[str], None], optional): The formula according to which build the model.
            In case of list of formula, the model matrix will be built in steps and
            concatenated together. Defaults to None.
        penalty_factor (Optional[List[float]], optional): the penalty used in the glmnet function in ipw. The penalty
            should have the same length as the formula list (and applies to each element of formula).
            Smaller penalty on some formula will lead to elements in that formula to get more adjusted, i.e. to have a higher chance to get into the model (and not zero out). A penalty of 0 will make sure the element is included in the model.
            If not provided, assume the same penalty (1) for all variables. Defaults to None.
        one_hot_encoding (bool, optional): whether to encode all factor variables in the model matrix with
            almost_one_hot_encoding. This is recomended in case of using
            LASSO on the data (Default: False).
            one_hot_encoding_greater_3 creates one-hot-encoding for all
            categorical variables with more than 2 categories (i.e. the
            number of columns will be equal to the number of categories),
            and only 1 column for variables with 2 levels (treatment contrast). Defaults to False.
        random_seed (int, optional): Random seed to use. Defaults to 2020.

    Raises:
        Exception: f"Sample indicator only has value {_n_unique}. This can happen when your sample or target are empty from unknown reason"
        NotImplementedError: if model is not "glmnet"

    Returns:
        Dict[str, Any]: A dictionary includes:
            "weight" --- The weights for the sample.
            "model" --- parameters of the model:fit, performance, X_matrix_columns, lambda,
                        weight_trimming_mean_ratio
            Shape of the Dict:
            {
                "weight": weights,
                "model": {
                    "method": "ipw",
                    "X_matrix_columns": X_matrix_columns_names,
                    "fit": fit,
                    "perf": performance,
                    "lambda": best_s,
                    "weight_trimming_mean_ratio": weight_trimming_mean_ratio,
                },
            }
    """
    logger.info("Starting ipw function")
    np.random.seed(
        random_seed
    )  # setting random seed for cases of variations in cvglmnet

    balance_util._check_weighting_methods_input(sample_df, sample_weights, "sample")
    balance_util._check_weighting_methods_input(target_df, target_weights, "target")

    variables = balance_util.choose_variables(sample_df, target_df, variables=variables)
    logger.debug(f"Join variables for sample and target: {variables}")

    sample_df = sample_df.loc[:, variables]
    target_df = target_df.loc[:, variables]

    if na_action == "drop":
        (sample_df, sample_weights) = balance_util.drop_na_rows(
            sample_df, sample_weights, "sample"
        )
        (target_df, target_weights) = balance_util.drop_na_rows(
            target_df, target_weights, "target"
        )
    sample_n = sample_df.shape[0]
    target_n = target_df.shape[0]

    # Applying transformations
    # Important! Variables that don't need transformations
    # should be transformed with the *identity function*,
    # otherwise will be dropped from the model
    sample_df, target_df = balance_adjustment.apply_transformations(
        (sample_df, target_df), transformations=transformations
    )
    variables = list(sample_df.columns)
    logger.debug(f"Final variables in the model: {variables}")

    logger.info("Building model matrix")
    model_matrix_output = balance_util.model_matrix(
        sample_df,
        target_df,
        variables,
        add_na=(na_action == "add_indicator"),
        return_type="one",
        return_var_type="sparse",
        # pyre-fixme[6]: for 7th parameter `formula` expected `Optional[List[str]]` but got `Union[None, List[str], str]`.
        # TODO: fix pyre issue
        formula=formula,
        penalty_factor=penalty_factor,
        one_hot_encoding=one_hot_encoding,
    )
    X_matrix = cast(
        Union[pd.DataFrame, np.ndarray, csc_matrix],
        model_matrix_output["model_matrix"],
    )
    X_matrix_columns_names = cast(
        List[str], model_matrix_output["model_matrix_columns_names"]
    )
    penalty_factor = cast(Optional[List[float]], model_matrix_output["penalty_factor"])
    # in cvglmnet: "penalty factors are internally rescaled to sum to nvars."
    # (https://glmnet-python.readthedocs.io/en/latest/glmnet_vignette.html)
    logger.info(
        f"The formula used to build the model matrix: {model_matrix_output['formula']}"
    )
    logger.info(f"The number of columns in the model matrix: {X_matrix.shape[1]}")
    logger.info(f"The number of rows in the model matrix: {X_matrix.shape[0]}")

    in_sample = np.concatenate((np.ones(sample_n), np.zeros(target_n)))
    _n_unique = np.unique(in_sample.reshape(in_sample.shape[0]))
    if len(_n_unique) == 1:
        raise Exception(
            f"Sample indicator only has value {_n_unique}. This can happen when your "
            "sample or target are empty from unknown reason"
        )

    if balance_classes:
        odds = np.sum(sample_weights) / np.sum(target_weights)
    else:
        odds = 1
    logger.debug(f"odds for balancing classes: {odds}")

    y0 = np.concatenate((np.zeros(sample_n), target_weights * odds))
    y1 = np.concatenate((sample_weights, np.zeros(target_n)))
    y = np.column_stack((y0, y1))
    # From glmnet documentation: https://glmnet-python.readthedocs.io/en/latest/glmnet_vignette.html
    # "For binomial logistic regression, the response variable y should be either
    # a factor with two levels, or a two-column matrix of counts or proportions."

    logger.debug(f"X_matrix shape: {X_matrix.shape}")
    logger.debug(f"y input shape: {y.shape}")
    logger.debug(
        f"penalty_factor frequency table {pd.crosstab(index=penalty_factor, columns='count')}"
        f"Note that these are normalized by cvglmnet"
    )

    if model == "glmnet":
        logger.info("Fitting logistic model")
        foldid = np.resize(range(10), y.shape[0])
        np.random.shuffle(
            foldid
        )  # shuffels the values of foldid - note that we set the seed in the beginning of the function, so this order is fixed
        logger.debug(
            f"foldid frequency table {pd.crosstab(index=foldid, columns='count')}"
        )
        logger.debug(f"first 10 elements of foldid: {foldid[0:9]}")
        with _patch_nan_in_amin_amax():
            # we use _patch_nan_in_amin_amax here since sometimes
            # cvglmnet could have one of the cross validated samples that
            # produce nan. In which case, the lambda search returns nan
            # instead of a value from the cross-validated options that successfully computed a lambda
            # The current monkey-patch solves this problem and makes the function fail less.
            with np.errstate(
                divide="ignore"
            ):  # ignoring np warning "divide by zero encountered in log"
                with _patch_scipy_random():
                    fit = cvglmnet(
                        x=X_matrix,
                        y=y,
                        family="binomial",
                        ptype="deviance",
                        alpha=1,
                        penalty_factor=penalty_factor,
                        nlambda=250,
                        lambda_min=np.array([1e-6]),
                        nfolds=10,
                        foldid=foldid,
                        maxit=5000,
                        *args,
                        **kwargs,
                    )
        logger.debug("Done with cvglmnet")
    else:
        raise NotImplementedError()
    logger.debug(f"fit['lambda_1se']: {fit['lambda_1se']}")

    X_matrix_sample = X_matrix[
        :sample_n,
    ].toarray()

    logger.info(f"max_de: {max_de}")
    if max_de is not None:
        regularisation_perf = choose_regularization(
            fit,
            sample_df,
            target_df,
            sample_weights,
            target_weights,
            X_matrix_sample,
            balance_classes,
            max_de,
        )
        best_s = regularisation_perf["best"]["s"]
        weight_trimming_mean_ratio = regularisation_perf["best"]["trim"]
        weight_trimming_percentile = None
    else:
        best_s = fit["lambda_1se"]
        regularisation_perf = None

    link = cvglmnetPredict(fit, X_matrix_sample, ptype="link", s=best_s)
    logger.debug("Predicting")
    weights = weights_from_link(
        link,
        balance_classes,
        sample_weights,
        target_weights,
        weight_trimming_mean_ratio,
        weight_trimming_percentile,
    )

    logger.info(f"Chosen lambda for cv: {best_s}")

    performance = cv_glmnet_performance(
        fit,
        feature_names=list(X_matrix_columns_names),
        s=best_s,
    )
    dev = performance["prop_dev_explained"]
    logger.info(f"Proportion null deviance explained {dev}")

    if np.unique(weights).shape[0] == 1:  # All weights are the same
        logger.warning("All weights are identical. The estimates will not be adjusted")

    coefs = performance["coefs"][1:]  # exclude intercept
    if all(abs(coefs) < 1e-14):
        # The value was determined by the unit-test test_adjustment/test_ipw_bad_adjustment_warnings
        logger.warning(
            (
                "All propensity model coefficients are zero, your covariates do "
                "not predict inclusion in the sample. The estimates will not be "
                "adjusted"
            )
        )

    if dev < 0.10:
        logger.warning(
            "The propensity model has low fraction null deviance explained "
            f"({dev}). Results may not be accurate"
        )

    out = {
        "weight": weights,
        "model": {
            "method": "ipw",
            "X_matrix_columns": X_matrix_columns_names,
            "fit": fit,
            "perf": performance,
            "lambda": best_s,
            "weight_trimming_mean_ratio": weight_trimming_mean_ratio,
        },
    }
    if max_de is not None:
        out["model"]["regularisation_perf"] = regularisation_perf

    logger.debug("Done ipw function")

    return out



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Dict, Union

import pandas as pd

logger: logging.Logger = logging.getLogger(__package__)


def adjust_null(
    sample_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_df: pd.DataFrame,
    target_weights: pd.Series,
    *args,
    **kwargs,
) -> Dict[str, Union[Dict[str, str], pd.Series]]:
    """Doesn't apply any adjustment to the data. Returns the design weights as they are.
    This may be useful when one needs the output of Sample.adjust() (i.e.: an adjusted object),
    but wishes to not run any model for it.

    Args:
        sample_df (pd.DataFrame): a dataframe representing the sample
        sample_weights (pd.Series): design weights for sample
        target_df (pd.DataFrame): a dataframe representing the target
        target_weights (pd.Series): design weights for target

    Returns:
        Dict[str, Union[Dict[str, str], pd.Series]]: Dict of weights (original sample weights) and model (with method = null_adjustment)
    """

    return {
        "weight": sample_weights,
        "model": {
            "method": "null_adjustment",
        },
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from typing import Callable, Dict, List, Literal, Tuple, Union

import numpy as np
import pandas as pd
import scipy

from balance import util as balance_util
from balance.weighting_methods import (
    adjust_null as balance_adjust_null,
    cbps as balance_cbps,
    ipw as balance_ipw,
    poststratify as balance_poststratify,
    rake as balance_rake,
)
from pandas.api.types import is_bool_dtype, is_numeric_dtype

logger: logging.Logger = logging.getLogger(__package__)


BALANCE_WEIGHTING_METHODS = {
    "ipw": balance_ipw.ipw,
    "cbps": balance_cbps.cbps,
    "null": balance_adjust_null.adjust_null,
    "poststratify": balance_poststratify.poststratify,
    "rake": balance_rake.rake,
}


def trim_weights(
    weights: Union[pd.Series, np.ndarray],
    # TODO: add support to more types of input weights? (e.g. list? other?)
    weight_trimming_mean_ratio: Union[float, int, None] = None,
    weight_trimming_percentile: Union[float, None] = None,
    verbose: bool = False,
    keep_sum_of_weights: bool = True,
) -> pd.Series:
    """Trim extreme weights.

    The user cannot supply both weight_trimming_mean_ratio and weight_trimming_percentile.
    If none are supplied, the original weights are returned.

    If weight_trimming_mean_ratio is not none, the weights are trimmed from above by
    mean(weights) * ratio. The weights are then normalized to have the original mean.
    Note that trimmed weights aren't actually bounded by trimming.ratio because the
    reduced weight is redistributed to arrive at the original mean

    Note that weight_trimming_percentile clips both sides of the distribution, unlike
    trimming that only trims the weights from above.
    For example, `weight_trimming_percentile=0.1` trims below the 10th percentile
    AND above the 90th. If you only want to trim the upper side, specify
    `weight_trimming_percentile = (0, 0.9)`.

    Args:
        weights (Union[pd.Series, np.ndarray]): pd.Series of weights to trim. np.ndarray will be turned into pd.Series) of weights.
        weight_trimming_mean_ratio (Union[float, int], optional): indicating the ratio from above according to which
            the weights are trimmed by mean(weights) * ratio. Defaults to None.
        weight_trimming_percentile (Union[float], optional): if weight_trimming_percentile is not none,
            then we apply winsorization using :func:`scipy.stats.mstats.winsorize`.
            See also: [https://en.wikipedia.org/wiki/Winsorizing].
            Defaults to None.
        verbose (bool, optional): whether to add to logger printout of trimming process.
            Defaults to False.
        keep_sum_of_weights (bool, optional): Set if the sum of weights after trimming
            should be the same as the sum of weights before trimming.
            Defaults to True.

    Raises:
        TypeError: If weights is not np.array or pd.Series.
        ValueError: If both weight_trimming_mean_ratio and weight_trimming_percentile are set.

    Returns:
        pd.Series (of type float64): Trimmed weights

     Examples:
        ::

            import pandas as pd
            from balance.adjustment import trim_weights
            print(trim_weights(pd.Series([1,2,3,40]), weight_trimming_mean_ratio = None))
               # 0     1.0
               # 1     2.0
               # 2     3.0
               # 3    40.0
               # dtype: float64

            print(trim_weights(pd.Series([1,2,3,40]), weight_trimming_mean_ratio = 2))
               # 0     1.586207
               # 1     3.172414
               # 2     4.758621
               # 3    36.482759
               # dtype: float64
    """

    if isinstance(weights, pd.Series):
        pass
    elif isinstance(weights, np.ndarray):
        weights = pd.Series(weights)
    else:
        raise TypeError(
            f"weights must be np.array or pd.Series, are of type: {type(weights)}"
        )

    if (weight_trimming_mean_ratio is not None) and (
        weight_trimming_percentile is not None
    ):
        raise ValueError(
            "Only one of weight_trimming_mean_ratio and "
            "weight_trimming_percentile can be set"
        )

    original_mean = np.mean(weights)

    if weight_trimming_mean_ratio is not None:
        max_val = weight_trimming_mean_ratio * original_mean
        percent_trimmed = weights[weights > max_val].count() / weights.count()
        weights = weights.clip(upper=max_val)
        if verbose:
            if percent_trimmed > 0:
                logger.debug("Clipping weights to %s (before renormalizing)" % max_val)
                logger.debug("Clipped %s of the weights" % percent_trimmed)
            else:
                logger.debug("No extreme weights were trimmed")
    elif weight_trimming_percentile is not None:
        # Winsorize
        weights = scipy.stats.mstats.winsorize(
            weights, limits=weight_trimming_percentile, inplace=False
        )
        if verbose:
            logger.debug(
                "Winsorizing weights to %s percentile" % str(weight_trimming_percentile)
            )

    if keep_sum_of_weights:
        weights = weights / np.mean(weights) * original_mean

    return weights


def default_transformations(
    dfs: Union[Tuple[pd.DataFrame, ...], List[pd.DataFrame]]
) -> Dict[str, Callable]:
    """
    Apply default transformations to dfs, i.e.
    quantize to numeric columns and fct_lump to non-numeric and boolean

    Args:
        dfs (Union[Tuple[pd.DataFrame, ...], List[pd.DataFrame]]): A list or tuple of dataframes

    Returns:
        Dict[str, Callable]: Dict of transformations
    """
    dtypes = {}
    for d in dfs:
        dtypes.update(d.dtypes.to_dict())

    transformations = {}
    for k, v in dtypes.items():
        # Notice that in pandas: pd.api.types.is_numeric_dtype(pd.Series([True, False])) == True
        # Hence, we need to explicitly check that not is_bool_dtype(v)
        # see: https://github.com/pandas-dev/pandas/issues/38378
        if (is_numeric_dtype(v)) and (not is_bool_dtype(v)):
            transformations[k] = balance_util.quantize
        else:
            transformations[k] = balance_util.fct_lump
    return transformations


def apply_transformations(
    dfs: Tuple[pd.DataFrame, ...],
    transformations: Union[Dict[str, Callable], str, None],
    drop: bool = True,
) -> Tuple[pd.DataFrame, ...]:
    """Apply the transformations specified in transformations to all of the dfs
        - if a column specified in `transformations` does not exist in the dataframes,
        it is added
        - if a column is not specified in `transformations`, it is dropped,
        unless drop==False
        - the dfs are concatenated together before transformations are applied,
        so functions like `max` are relative to the column in all dfs
        - Cannot transform the same variable twice, or add a variable and then transform it
        (i.e. the definition of the added variable should include the transformation)
        - if you get a cryptic error about mismatched data types, make sure your
        transformations are not being treated as additions because of missing
        columns (use `_set_warnings("DEBUG")` to check)

    Args:
        dfs (Tuple[pd.DataFrame, ...]): The DataFrames on which to operate
        transformations (Union[Dict[str, Callable], str, None]): Mapping from column name to function to apply.
            Transformations of existing columns should be specified as functions
            of those columns (e.g. `lambda x: x*2`), whereas additions of new
            columns should be specified as functions of the DataFrame
            (e.g. `lambda x: x.column_a + x.column_b`).
        drop (bool, optional): Whether to drop columns which are
              not specified in `transformations`. Defaults to True.

    Raises:
        NotImplementedError: When passing an unknown "transformations" argument.

    Returns:
        Tuple[pd.DataFrame, ...]: tuple of pd.DataFrames

    Examples:
        ::

            from balance.adjustment import apply_transformations
            import pandas as pd
            import numpy as np

            apply_transformations(
                (pd.DataFrame({'d': [1, 2, 3], 'e': [4, 5, 6]}),),
                {'d': lambda x: x*2, 'f': lambda x: x.d+x.e}
            )

                # (   f  d
                #  0  5  2
                #  1  7  4
                #  2  9  6,)

    """
    # TODO: change assert to raise
    assert isinstance(dfs, tuple), "'dfs' argument must be a tuple of DataFrames"
    assert all(
        isinstance(x, pd.DataFrame) for x in dfs
    ), "'dfs' must contain DataFrames"

    if transformations is None:
        return dfs
    elif isinstance(transformations, str):
        if transformations == "default":
            transformations = default_transformations(dfs)
        else:
            raise NotImplementedError(f"Unknown transformations {transformations}")

    ns = [0] + list(np.cumsum([x.shape[0] for x in dfs]))
    boundaries = [(ns[i], ns[i + 1]) for i in range(0, len(ns) - 1)]
    indices = [x.index for x in dfs]

    all_data = pd.concat(dfs).reset_index(drop=True)
    # This is to avoid issues with trnasformations that cannot
    # be done on object with duplicate indecies

    # additions is new columns to add to data. i.e.: column names that appear in transformations
    #   but are not present in all_data.
    additions = {k: v for k, v in transformations.items() if k not in all_data.columns}
    transformations = {
        k: v for k, v in transformations.items() if k in all_data.columns
    }
    logger.info(f"Adding the variables: {list(additions.keys())}")
    logger.info(f"Transforming the variables: {list(transformations.keys())}")
    logger.debug(
        f"Total number of added or transformed variables: {len(additions) + len(transformations)}"
    )

    # TODO: change assert to raise
    assert (
        len(additions) + len(transformations)
    ) > 0, "No transformations or additions passed"

    if len(additions) > 0:
        added = all_data.assign(**additions).loc[:, list(additions.keys())]
    else:
        added = None

    if len(transformations) > 0:
        # NOTE: .copy(deep=False) is used to avoid a false alarm that sometimes happen.
        # When we take a slice of the DataFrame (all_data[k]), it is passed to a function
        # inside v from transformations (e.g.: fct_lump), it would then sometimes raise:
        # SettingWithCopyWarning: A value is trying to be set on a copy of a slice from a DataFrame.
        # Adding .copy(deep=False) solves this.
        # See: https://stackoverflow.com/a/54914752
        transformed = pd.DataFrame(
            {k: v(all_data.copy(deep=False)[k]) for k, v in transformations.items()}
        )
    else:
        transformed = None

    out = pd.concat((added, transformed), axis=1)

    dropped_columns = list(set(all_data.columns.values) - set(out.columns.values))

    if len(dropped_columns) > 0:
        if drop:
            logger.warning(f"Dropping the variables: {dropped_columns}")
        else:
            out = pd.concat((out, all_data.loc[:, dropped_columns]), axis=1)
    logger.info(f"Final variables in output: {list(out.columns)}")

    for column in out:
        logger.debug(
            f"Frequency table of column {column}:\n{out[column].value_counts(dropna=False)}"
        )
        logger.debug(
            f"Number of levels of column {column}:\n{out[column].nunique(dropna=False)}"
        )

    res = tuple(out[i:j] for (i, j) in boundaries)
    res = tuple(x.set_index(i) for x, i in zip(res, indices))

    return res


def _find_adjustment_method(
    method: Literal["cbps", "ipw", "null", "poststratify", "rake"],
    WEIGHTING_METHODS: Dict[str, Callable] = BALANCE_WEIGHTING_METHODS,
) -> Callable:
    """This function translates a string method argument to the function itself.

    Args:
        method (Literal["cbps", "ipw", "null", "poststratify", "rake"]): method for adjustment: cbps, ipw, null, poststratify
        WEIGHTING_METHODS (Dict[str, Callable]): A dict where keys are strings of function names, and the values are
            the functions themselves.

    Returns:
        Callable: The function for adjustment
    """
    if method in WEIGHTING_METHODS.keys():
        adjustment_function = WEIGHTING_METHODS[method]
    else:
        raise ValueError(f"Unknown adjustment method: '{method}'")

    return adjustment_function



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import random

import balance

import balance.testutil

import numpy as np
import pandas as pd

from balance.adjustment import (
    apply_transformations,
    default_transformations,
    trim_weights,
)
from balance.sample_class import Sample
from balance.util import fct_lump, quantize
from balance.weighting_methods import (
    cbps as balance_cbps,
    ipw as balance_ipw,
    poststratify as balance_poststratify,
)

EPSILON = 0.00001


sample = Sample.from_frame(
    df=pd.DataFrame(
        {
            "a": (1, 2, 3, 1),
            "b": (-42, 8, 2, -42),
            "o": (7, 8, 9, 10),
            "c": ("x", "y", "z", "x"),
            "id": (1, 2, 3, 4),
            "w": (0.5, 2, 1, 1),
        }
    ),
    id_column="id",
    weight_column="w",
    outcome_columns="o",
)

target = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3),
            "b": (-42, 8, 2),
            "c": ("x", "y", "z"),
            "id": (1, 2, 3),
            "w": (2, 0.5, 1),
        }
    ),
    id_column="id",
    weight_column="w",
)


class TestAdjustment(
    balance.testutil.BalanceTestCase,
):
    def test_trim_weights(self):

        # Test no trimming
        # Notice how it changes the dtype of int64 to float64~
        pd.testing.assert_series_equal(
            trim_weights(pd.Series([0, 1, 2])), pd.Series([0.0, 1.0, 2.0])
        )
        self.assertEqual(type(trim_weights(pd.Series([0, 1, 2]))), pd.Series)
        self.assertEqual(trim_weights(pd.Series([0, 1, 2])).dtype, np.float64)

        random.seed(42)
        w = np.random.uniform(0, 1, 10000)
        self.assertEqual(
            trim_weights(
                w,
                weight_trimming_percentile=None,
                weight_trimming_mean_ratio=None,
                keep_sum_of_weights=False,
            ),
            w,
        )

        # Test exceptions
        with self.assertRaisesRegex(
            TypeError, "weights must be np.array or pd.Series, are of type*"
        ):
            trim_weights("Strings don't get trimmed", weight_trimming_mean_ratio=1)
        with self.assertRaisesRegex(ValueError, "Only one"):
            trim_weights(
                np.array([0, 1, 2]),
                1,
                1,
            )

        # Test weight_trimming_mean_ratio
        random.seed(42)
        w = np.random.uniform(0, 1, 10000)
        res = trim_weights(w, weight_trimming_mean_ratio=1)
        self.assertAlmostEqual(np.mean(w), np.mean(res), delta=EPSILON)
        self.assertAlmostEqual(
            np.mean(w) / np.min(w), np.max(res) / np.min(res), delta=EPSILON
        )

        # Test weight_trimming_percentile
        random.seed(42)
        w = np.random.uniform(0, 1, 10000)
        self.assertTrue(
            max(
                trim_weights(
                    w, weight_trimming_percentile=(0, 0.11), keep_sum_of_weights=False
                )
            )
            < 0.9
        )
        self.assertTrue(
            min(
                trim_weights(
                    w, weight_trimming_percentile=(0.11, 0), keep_sum_of_weights=False
                )
            )
            > 0.1
        )
        e = trim_weights(w, weight_trimming_percentile=(0.11, 0.11))
        self.assertTrue(min(e) > 0.1)
        self.assertTrue(max(e) < 0.9)

    def test_default_transformations(self):
        # For multiple dataframes
        input = (
            pd.DataFrame({"a": (1, 2), "b": ("a", "b")}),
            pd.DataFrame({"c": (1, 2), "d": ("a", "b")}),
        )
        r = default_transformations(input)
        self.assertEqual(
            r,
            {
                "a": quantize,
                "b": fct_lump,
                "c": quantize,
                "d": fct_lump,
            },
        )

        # For one dataframe
        input = pd.DataFrame({"a": (1, 2), "b": ("a", "b")})
        r = default_transformations([input])
        self.assertEqual(
            r,
            {
                "a": quantize,
                "b": fct_lump,
            },
        )

        # For boolean and Int64 input
        input = pd.DataFrame({"a": (1, 2), "b": (True, False)})
        input = input.astype(
            dtype={
                "a": "Int64",
                "b": "boolean",
            }
        )
        r = default_transformations([input])
        self.assertEqual(
            r,
            {
                "a": quantize,
                "b": fct_lump,
            },
        )

    def test_default_transformations_pd_int64(self):
        nullable_int = pd.DataFrame({"a": pd.array((1, 2), dtype="Int64")})

        numpy_int = nullable_int.astype(np.int64)

        test = default_transformations([nullable_int])
        truth = default_transformations([numpy_int])

        self.assertEqual(test, truth)

    def test_apply_transformations(self):
        s = pd.DataFrame({"d": [1, 2, 3], "e": [1, 2, 3]})
        t = pd.DataFrame({"d": [4, 5, 6, 7], "e": [1, 2, 3, 4]})

        transformations = {"d": lambda x: x * 2, "f": lambda x: x.d + 1}
        r = apply_transformations((s, t), transformations)

        e = (
            pd.DataFrame({"d": [2, 4, 6], "f": [2, 3, 4]}),
            pd.DataFrame({"d": [8, 10, 12, 14], "f": [5, 6, 7, 8]}),
        )

        self.assertEqual(r[0], e[0], lazy=True)
        self.assertEqual(r[1], e[1], lazy=True)

        # No transformations or additions
        self.assertEqual(apply_transformations((s, t), None), (s, t))

        # Only transformations
        r = apply_transformations((s, t), {"d": lambda x: x * 2})
        e = (pd.DataFrame({"d": [2, 4, 6]}), pd.DataFrame({"d": [8, 10, 12, 14]}))
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Only additions
        r = apply_transformations((s, t), {"f": lambda x: x.d + 1})
        e = (pd.DataFrame({"f": [2, 3, 4]}), pd.DataFrame({"f": [5, 6, 7, 8]}))
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Warns about dropping variable
        self.assertWarnsRegexp(
            r"Dropping the variables: \['e'\]",
            apply_transformations,
            (s, t),
            transformations,
        )

        # Does not drop
        r = apply_transformations((s, t), transformations, drop=False)
        e = (
            pd.DataFrame({"d": [2, 4, 6], "e": [1, 2, 3], "f": [2, 3, 4]}),
            pd.DataFrame({"d": [8, 10, 12, 14], "e": [1, 2, 3, 4], "f": [5, 6, 7, 8]}),
        )
        self.assertEqual(r[0], e[0], lazy=True)
        self.assertEqual(r[1], e[1], lazy=True)

        # Works on three dfs
        q = pd.DataFrame({"d": [8, 9], "g": [1, 2]})
        r = apply_transformations((s, t, q), transformations)
        e = (
            pd.DataFrame({"d": [2, 4, 6], "f": [2, 3, 4]}),
            pd.DataFrame({"d": [8, 10, 12, 14], "f": [5, 6, 7, 8]}),
            pd.DataFrame({"d": [16, 18], "f": [9, 10]}),
        )
        self.assertEqual(r[0], e[0], lazy=True)
        self.assertEqual(r[1], e[1], lazy=True)
        self.assertEqual(r[2], e[2], lazy=True)

        # Test that functions are computed over all dfs passed, not each individually
        transformations = {"d": lambda x: x / max(x)}
        r = apply_transformations((s, t), transformations)
        e = (
            pd.DataFrame({"d": [2 / 14, 4 / 14, 6 / 14]}),
            pd.DataFrame({"d": [8 / 14, 10 / 14, 12 / 14, 14 / 14]}),
        )
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Transformation of a column which does not exist in one of the dataframes
        s = pd.DataFrame({"d": [1, 2, 3], "e": [1, 2, 3]})
        t = pd.DataFrame({"d": [4, 5, 6, 7]})
        transformations = {"e": lambda x: x * 2}
        r = apply_transformations((s, t), transformations)
        e = (
            pd.DataFrame({"e": [2.0, 4.0, 6.0]}),
            pd.DataFrame({"e": [np.nan, np.nan, np.nan, np.nan]}),
        )
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Additon of a column based on one which does not exist in one of the dataframes
        transformations = {"f": lambda x: x.e * 2}
        r = apply_transformations((s, t), transformations)
        e = (
            pd.DataFrame({"f": [2.0, 4.0, 6.0]}),
            pd.DataFrame({"f": [np.nan, np.nan, np.nan, np.nan]}),
        )
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Column which does not exist in one of the dataframes
        # and is also specified
        s = pd.DataFrame({"d": [1, 2, 3], "e": [0, 0, 0]})
        t = pd.DataFrame({"d": [4, 5, 6, 7]})
        transformations = {"e": lambda x: x + 1}
        r = apply_transformations((s, t), transformations)
        e = (
            pd.DataFrame({"e": [1.0, 1.0, 1.0]}),
            pd.DataFrame({"e": [np.nan, np.nan, np.nan, np.nan]}),
        )
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Test that indices are ignored in splitting dfs
        s = pd.DataFrame({"d": [1, 2, 3]}, index=(5, 6, 7))
        t = pd.DataFrame({"d": [4, 5, 6, 7]}, index=(0, 1, 2, 3))
        transformations = {"d": lambda x: x}
        r = apply_transformations((s, t), transformations)
        e = (s, t)
        self.assertEqual(r[0], e[0])
        self.assertEqual(r[1], e[1])

        # Test indices are handeled okay (this example reuired reset_index of all_data)
        s = pd.DataFrame({"a": (0, 0, 0, 0, 0, 0, 0, 0)})
        t = pd.DataFrame({"a": (1, 1, 1, 1)})
        r = apply_transformations((s, t), "default")
        e = (
            pd.DataFrame({"a": ["(-0.001, 0.7]"] * 8}),
            pd.DataFrame({"a": ["(0.7, 1.0]"] * 4}),
        )
        self.assertEqual(r[0].astype(str), e[0])
        self.assertEqual(r[1].astype(str), e[1])

        #  Test default transformations
        s = pd.DataFrame({"d": range(0, 100), "e": ["a"] * 96 + ["b"] * 4})
        t = pd.DataFrame({"d": range(0, 100), "e": ["a"] * 96 + ["b"] * 4})
        r_s, r_t = apply_transformations((s, t), "default")

        self.assertEqual(r_s["d"].drop_duplicates().values.shape[0], 10)
        self.assertEqual(r_t["d"].drop_duplicates().values.shape[0], 10)

        self.assertEqual(r_s["e"].drop_duplicates().values, ("a", "_lumped_other"))
        self.assertEqual(r_t["e"].drop_duplicates().values, ("a", "_lumped_other"))

    def test_invalid_input_to_apply_transformations(self):
        # Test non-existent transformation
        self.assertRaisesRegex(
            NotImplementedError,
            "Unknown transformations",
            apply_transformations,
            (sample.df,),
            "foobar",
        )

        # Test non-dataframe input
        self.assertRaisesRegex(
            AssertionError,
            "'dfs' must contain DataFrames",
            apply_transformations,
            (sample,),
            "foobar",
        )

        # Test non-tuple input
        self.assertRaisesRegex(
            AssertionError,
            "'dfs' argument must be a tuple of DataFrames",
            apply_transformations,
            sample.df,
            "foobar",
        )

    def test__find_adjustment_method(self):
        self.assertTrue(
            balance.adjustment._find_adjustment_method("ipw") is balance_ipw.ipw
        )
        self.assertTrue(
            balance.adjustment._find_adjustment_method("cbps") is balance_cbps.cbps
        )
        self.assertTrue(
            balance.adjustment._find_adjustment_method("poststratify")
            is balance_poststratify.poststratify
        )
        with self.assertRaisesRegex(ValueError, "Unknown adjustment method*"):
            balance.adjustment._find_adjustment_method("some_other_value")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

# LICENSE file in the root directory of this source tree.

from pathlib import Path
from typing import AnyStr, IO, Union

FilePathOrBuffer = Union[str, Path, IO[AnyStr]]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import io
import re
import sys

import unittest
from contextlib import contextmanager
from typing import Any, Union

import numpy as np
import pandas as pd


def _assert_frame_equal_lazy(
    x: pd.DataFrame, y: pd.DataFrame, lazy: bool = True
) -> None:
    """Wrapper around pd.testing.assert_frame_equal, which transforms the
    dataframes to ignore some errors.

    Ignores order of columns

    Args:
        x (pd.DataFrame): DataFrame to compare
        y (pd.DataFrame): DataFrame to compare
        lazy (bool, optional): Should Ignores be applied. Defaults to True.

    Returns:
        None.
    """
    if lazy:
        x = x.sort_index(axis=0).sort_index(axis=1)
        y = y.sort_index(axis=0).sort_index(axis=1)

    return pd.testing.assert_frame_equal(x, y)


def _assert_index_equal_lazy(x: pd.Index, y: pd.Index, lazy: bool = True) -> None:
    """
    Wrapper around pd.testing.assert_index_equal which transforms the
    dataindexs to ignore some errors.

    Ignores:
        - order of entries

    Args:
        x (pd.Index): Index to compare
        y (pd.Index): Index to compare
        lazy (bool, optional): Should Ignores be applied. Defaults to True.
    """
    if lazy:
        x = x.sort_values()
        y = y.sort_values()

    return pd.testing.assert_index_equal(x, y)


@contextmanager
def _capture_output():
    redirect_out, redirect_err = io.StringIO(), io.StringIO()
    original_out, original_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = redirect_out, redirect_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = original_out, original_err


class BalanceTestCase(unittest.TestCase):
    # Some Warns
    def assertIfWarns(self, callable, *args, **kwargs) -> None:
        with self.assertLogs(level="NOTSET") as cm:
            callable(*args, **kwargs)
            self.assertTrue(len(cm.output) > 0, "No warning produced.")

    def assertNotWarns(self, callable, *args, **kwargs) -> None:
        output = None
        try:
            with self.assertLogs() as cm:
                callable(*args, **kwargs)
                output = cm
        except AssertionError:
            return
        raise AssertionError(f"Warning produced {output.output}.")

    def assertWarnsRegexp(self, regexp, callable, *args, **kwargs) -> None:
        with self.assertLogs(level="NOTSET") as cm:
            callable(*args, **kwargs)
            self.assertTrue(
                any((re.search(regexp, c) is not None) for c in cm.output),
                f"Warning {cm.output} does not match regex {regexp}.",
            )

    def assertNotWarnsRegexp(self, regexp, callable, *args, **kwargs) -> None:
        with self.assertLogs(level="NOTSET") as cm:
            callable(*args, **kwargs)
            self.assertFalse(
                any((re.search(regexp, c) is not None) for c in cm.output),
                f"Warning {cm.output} matches regex {regexp}.",
            )

    # Some Equal
    def assertEqual(
        self,
        first: Union[np.ndarray, pd.DataFrame, pd.Index, pd.Series, Any],
        second: Union[np.ndarray, pd.DataFrame, pd.Index, pd.Series, Any],
        msg: Any = ...,
        **kwargs,
    ) -> None:
        """
        Check if first and second are equal.
        Uses np.testing.assert_array_equal for np.ndarray,
        _assert_frame_equal_lazy for pd.DataFrame,
        assert_series_equal for  pd.DataFrame,
        _assert_index_equal_lazy for pd.Index,
        or unittest.TestCase.assertEqual otherwise.

        Args:
            first (Union[np.ndarray, pd.DataFrame, pd.Index, pd.Series]): first element to compare.
            second (Union[np.ndarray, pd.DataFrame, pd.Index, pd.Series]): second element to compare.
            msg (Any, optional): The error message on failure.
        """
        lazy: bool = kwargs.get("lazy", False)
        if isinstance(first, np.ndarray) or isinstance(second, np.ndarray):
            np.testing.assert_array_equal(first, second, **kwargs)
        elif isinstance(first, pd.DataFrame) or isinstance(second, pd.DataFrame):
            _assert_frame_equal_lazy(
                first,
                second,
                lazy,
            )
        elif isinstance(first, pd.Series) or isinstance(second, pd.Series):
            pd.testing.assert_series_equal(first, second)
        elif isinstance(first, pd.Index) or isinstance(second, pd.Index):
            _assert_index_equal_lazy(first, second, lazy)
        else:
            super().assertEqual(first, second, msg=msg, **kwargs)

    # Some Prints
    def assertPrints(self, callable, *args, **kwargs) -> None:
        with _capture_output() as (out, err):
            callable(*args, **kwargs)
            out, err = out.getvalue(), err.getvalue()
            self.assertTrue((len(out) + len(err)) > 0, "No printed output.")

    def assertNotPrints(self, callable, *args, **kwargs) -> None:
        with _capture_output() as (out, err):
            callable(*args, **kwargs)
            out, err = out.getvalue(), err.getvalue()
            self.assertTrue(
                (len(out) + len(err)) == 0,
                f"Printed output is longer than 0: {(out, err)}.",
            )

    def assertPrintsRegexp(self, regexp, callable, *args, **kwargs) -> None:
        with _capture_output() as (out, err):
            callable(*args, **kwargs)
            out, err = out.getvalue(), err.getvalue()
            self.assertTrue(
                any((re.search(regexp, o) is not None) for o in (out, err)),
                f"Printed output {(out, err)} does not match regex {regexp}.",
            )

    def assertNotPrintsRegexp(self, regexp, callable, *args, **kwargs) -> None:
        with _capture_output() as (out, err):
            callable(*args, **kwargs)
            out, err = out.getvalue(), err.getvalue()
            self.assertFalse(
                any((re.search(regexp, o) is not None) for o in (out, err)),
                f"Printed output {(out, err)} matches regex {regexp}.",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import logging
import sys

import balance.testutil

import numpy as np
import pandas as pd


class TestTestUtil(
    balance.testutil.BalanceTestCase,
):
    def test_testutil(self):
        # _assert_frame_equal_lazy
        self.assertRaises(
            AssertionError,
            balance.testutil._assert_frame_equal_lazy,
            pd.DataFrame({"a": (1, 2, 3)}),
            pd.DataFrame({"a": (1, 2, 4)}),
        )

        self.assertRaises(
            AssertionError,
            balance.testutil._assert_frame_equal_lazy,
            pd.DataFrame({"a": (1, 2, 3), "b": (1, 2, 3)}),
            pd.DataFrame({"a": (1, 2, 4), "c": (1, 2, 3)}),
        )

        a = pd.DataFrame({"a": (1, 2, 3), "b": (4, 5, 6)})
        b = pd.DataFrame({"a": (1, 2, 3), "b": (4, 5, 6)}, columns=("b", "a"))

        # Doesn't raise an error
        balance.testutil._assert_frame_equal_lazy(a, b)

        self.assertRaises(
            AssertionError, balance.testutil._assert_frame_equal_lazy, a, b, False
        )

        # _assert_index_equal_lazy
        self.assertRaises(
            AssertionError,
            balance.testutil._assert_index_equal_lazy,
            pd.Index([1, 2, 3]),
            pd.Index([1, 2, 4]),
        )

        a = pd.Index([1, 2, 3])
        b = pd.Index([1, 3, 2])
        # Doesn't raise an error
        balance.testutil._assert_index_equal_lazy(a, b)
        self.assertRaises(
            AssertionError, balance.testutil._assert_index_equal_lazy, a, b, False
        )


class TestTestUtil_BalanceTestCase_Equal(
    balance.testutil.BalanceTestCase,
):
    def test_additional_equality_tests_mixin(self):
        # np.array
        self.assertRaises(AssertionError, self.assertEqual, 1, 2)
        self.assertRaises(
            AssertionError, self.assertEqual, np.array((1, 2)), np.array((2, 1))
        )

        # Does not raise
        self.assertEqual(1, 1)
        self.assertEqual(np.array((1, 2)), np.array((1, 2)))

        # pd.DataFrames
        # The default is for non-lazy testing of pandas DataFrames
        a = pd.DataFrame({"a": (1, 2, 3), "b": (4, 5, 6)})
        b = pd.DataFrame({"a": (1, 2, 3), "b": (4, 5, 6)}, columns=("b", "a"))

        # Does raise an error by default or if lazy=False
        self.assertRaises(AssertionError, self.assertEqual, a, b)
        self.assertRaises(AssertionError, self.assertEqual, a, b, lazy=False)

        # Doesn't raise an error
        self.assertEqual(a, b, lazy=True)

        # pd.Series
        self.assertEqual(pd.Series([1, 2]), pd.Series([1, 2]))
        self.assertRaises(
            AssertionError, self.assertEqual, pd.Series([1, 2]), pd.Series([2, 1])
        )

        #  Indices
        self.assertEqual(pd.Index((1, 2)), pd.Index((1, 2)))
        self.assertRaises(
            AssertionError, self.assertEqual, pd.Index((1, 2)), pd.Index((2, 1))
        )
        self.assertRaises(
            AssertionError,
            self.assertEqual,
            pd.Index((1, 2)),
            pd.Index((2, 1)),
            lazy=False,
        )
        self.assertEqual(pd.Index((1, 2)), pd.Index((2, 1)), lazy=True)

        # Other types
        self.assertEqual("a", "a")
        self.assertRaises(AssertionError, self.assertEqual, "a", "b")


class TestTestUtil_BalanceTestCase_Warns(
    balance.testutil.BalanceTestCase,
):
    def test_unit_test_warning_mixin(self):
        logger = logging.getLogger(__package__)

        self.assertIfWarns(lambda: logger.warning("test"))
        self.assertNotWarns(lambda: "x")

        self.assertWarnsRegexp("abc", lambda: logger.warning("abcde"))
        self.assertRaises(
            AssertionError,
            self.assertWarnsRegexp,
            "abcdef",
            lambda: logger.warning("abcde"),
        )

        self.assertNotWarnsRegexp("abcdef", lambda: logger.warning("abcde"))


class TestTestUtil_BalanceTestCase_Print(
    balance.testutil.BalanceTestCase,
):
    def test_unit_test_print_mixin(self):
        self.assertPrints(lambda: print("x"))
        self.assertNotPrints(lambda: "x")

        self.assertPrintsRegexp("abc", lambda: print("abcde"))
        self.assertRaises(
            AssertionError, self.assertPrintsRegexp, "abcdef", lambda: print("abcde")
        )

        # assertPrintsRegexp() doesn't necessarily work with logging.warning(),
        # as logging handlers can change (e.g. in PyTest)
        self.assertPrintsRegexp("abc", lambda: print("abcde", file=sys.stderr))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import logging
import re
from typing import List, Literal, Union

import numpy as np
import pandas as pd
from balance.stats_and_plots.weighted_stats import (
    descriptive_stats,
    weighted_mean,
    weighted_var,
)
from balance.stats_and_plots.weights_stats import _check_weights_are_valid

logger: logging.Logger = logging.getLogger(__package__)

# TODO: add?
# from scipy.stats import wasserstein_distance

##########################################
# Weighted comparisons - functions to compare one or two data sources with one or two sources of weights
##########################################


# TODO: fix the r_indicator function. The current implementation is broken since it
#       seems to wrongly estimate N.
#       This seems to attempt to reproduce equation 2.2.2, in page 5 in
#       "Indicators for the representativeness of survey response"
#       by Jelke Bethlehem, Fannie Cobben, and Barry Schouten
#       See pdf: https://www150.statcan.gc.ca/n1/en/pub/11-522-x/2008000/article/10976-eng.pdf?st=Zi4d4zld
#       From: https://www150.statcan.gc.ca/n1/pub/11-522-x/2008000/article/10976-eng.pdf
# def r_indicator(sample_p: np.float64, target_p: np.float64) -> np.float64:
#     p = np.concatenate((sample_p, target_p))
#     N = len(sample_p) + len(target_p)
#     return 1 - 2 * np.sqrt(1 / (N - 1) * np.sum((p - np.mean(p)) ** 2))


def _weights_per_covars_names(covar_names: List) -> pd.DataFrame:
    # TODO (p2): consider if to give weights that are proportional to the proportion of this covar in the population
    #           E.g.: if merging varios platforms, maybe if something like windows has very few users, it's impact on the ASMD
    #           should be smaller (similar to how kld works).
    #           The current function structure won't support it, and this would require some extra input (i.e.: baseline target pop proportion)
    """
    Figure out how much weight to give to each column name for ASMD averaging.
    This is meant for post-processing df produced from model_matrix that include
    one-hot categorical variables. This function helps to resolve the needed weight to add
    to columns after they are broken down by the one-hot encoding used in model_matrix.
    Each of these will count for 1/num_levels.
    It's OK to assume that variables with '[]' in the name are one-hot
    categoricals because model_matrix enforces it.

    Args:
        covar_names (List): A list with names of covariate.

    Returns:
        pd.DataFrame: with two columns, one for weights and another for main_covar_names,
        with rows for each of the columns from 'covar_names'

    Examples:
        ::

            asmd_df = pd.DataFrame(
            {
                'age': 0.5,
                'education[T.high_school]': 1,
                'education[T.bachelor]': 1,
                'education[T.masters]': 1,
                'education[T.phd]': 1,
            }, index = ('self', ))

            input = asmd_df.columns.values.tolist()
                # input
                # ['age',
                #  'education[T.high_school]',
                #  'education[T. bachelor]',
                #  'education[T. masters]',
                #  'education[T. phd]']

            _weights_per_covars_names(input).to_dict()
                # Output:
                # {'weight': {'age': 1.0,
                #   'education[T.high_school]': 0.25,
                #   'education[T.bachelor]': 0.25,
                #   'education[T.masters]': 0.25,
                #   'education[T.phd]': 0.25},
                #  'main_covar_names': {'age': 'age',
                #   'education[T.high_school]': 'education',
                #   'education[T.bachelor]': 'education',
                #   'education[T.masters]': 'education',
                #   'education[T.phd]': 'education'}}
    """
    columns_to_original_variable = {v: re.sub(r"\[.*\]$", "", v) for v in covar_names}
    counts = collections.Counter(columns_to_original_variable.values())
    weights = pd.DataFrame(
        {k: 1 / counts[v] for k, v in columns_to_original_variable.items()},
        index=("weight",),
    )
    _check_weights_are_valid(weights)  # verify nothing odd has occured.
    main_covar_names = pd.DataFrame.from_dict(
        columns_to_original_variable,
        orient="index",
        columns=[
            "main_covar_names",
        ],
    )
    return pd.concat([weights.transpose(), main_covar_names], axis=1)


# TODO: add memoization
def asmd(
    sample_df: pd.DataFrame,
    target_df: pd.DataFrame,
    sample_weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    target_weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    std_type: Literal["target", "sample", "pooled"] = "target",
    aggregate_by_main_covar: bool = False,
) -> pd.Series:
    """
    Calculate the Absolute Standardized Mean Deviation (ASMD, a.k.a
    absolute standardized mean difference) between the columns of two DataFrames
    (or BalanceDFs).
    It uses weighted average and std for the calculations.
    This is the same as taking the absolute value of Cohen's d statistic,
    with a specific choice of the standard deviation (std).

    https://en.wikipedia.org/wiki/Effect_size#Cohen's_d

    As opposed to Cohen's d, the current asmd implementation has several options of std calculation,
    see the arguments section for details.

    Unlike in R package {cobalt}, in the current implementation of asmd:
    - the absolute value is taken
    - un-represented levels of categorical variables are treated as missing,
      not as 0.
    - differences for categorical variables are also weighted by default

    The function omits columns that starts with the name "_is_na_"

    If column names of sample_df and target_df are different, it will only calculate asmd for
    the overlapping columns. The rest will be np.nan.
    The mean(asmd) will be calculated while treating the nan values as 0s.

    Args:
        sample_df (pd.DataFrame): source group of the asmd comparison
        target_df (pd.DataFrame): target group of the asmd comparison.
            The column names should be the same as the ones from sample_df.
        sample_weights (Union[ List, pd.Series, np.ndarray, ], optional): weights to use.
            The default is None.
            If no weights are passed (None), it will use an array of 1s.
        target_weights (Union[ List, pd.Series, np.ndarray, ], optional): weights to use.
            The default is None.
            If no weights are passed (None), it will use an array of 1s.
        std_type (Literal["target", "sample", "pooled"], optional):
            How the standard deviation should be calculated.
            The options are: "target", "sample" and "pooled". Defaults to "target".
            "target" means we use the std from the target population.
            "sample" means we use the std from the sample population.
            "pooled" means we use the simple arithmetic average of
                the variance from the sample and target population.
                We then take the square root of that value to get the pooled std.
                Notice that this is done while giving the same weight to both sources.
                I.e.: there is NO over/under weighting sample or target based on their
                respective sample sizes (in contrast to how pooled sd is calculated in
                Cohen's d statistic).
        aggregate_by_main_covar (bool):
            If to use :func:`_aggregate_asmd_by_main_covar` to aggregate (average) the asmd based on variable name.
            Default is False.
    Returns:
        pd.Series: a Series indexed on the names of the columns in the input DataFrames.
        The values (of type np.float64) are of the ASMD calculation.
        The last element is 'mean(asmd)', which is the average of the calculated ASMD values.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots import weighted_comparisons_stats

            a1 = pd.Series((1, 2))
            b1 = pd.Series((-1, 1))
            a2 = pd.Series((3, 4))
            b2 = pd.Series((-2, 2))
            w1 = pd.Series((1, 1))
            w2 = w1

            r = weighted_comparisons_stats.asmd(
                        pd.DataFrame({"a": a1, "b": b1}),
                        pd.DataFrame({"a": a2, "b": b2}),
                        w1,
                        w2,
                    )

            exp_a = np.abs(a1.mean() - a2.mean()) / a2.std()
            exp_b = np.abs(b1.mean() - b2.mean()) / b2.std()
            print(r)
            print(exp_a)
            print(exp_b)

            # output:
            # {'a': 2.82842712474619, 'b': 0.0, 'mean(asmd)': 1.414213562373095}
            # 2.82842712474619
            # 0.0



            a1 = pd.Series((1, 2))
            b1_A = pd.Series((1, 3))
            b1_B = pd.Series((-1, -3))
            a2 = pd.Series((3, 4))
            b2_A = pd.Series((2, 3))
            b2_B = pd.Series((-2, -3))
            w1 = pd.Series((1, 1))
            w2 = w1

            r = weighted_comparisons_stats.asmd(
                pd.DataFrame({"a": a1, "b[A]": b1_A, "b[B]": b1_B}),
                pd.DataFrame({"a": a2, "b[A]": b2_A, "b[B]": b2_B}),
                w1,
                w2,
            ).to_dict()

            print(r)
            # {'a': 2.82842712474619, 'b[A]': 0.7071067811865475, 'b[B]': 0.7071067811865475, 'mean(asmd)': 1.7677669529663689}

            # Check that using aggregate_by_main_covar works
            r = weighted_comparisons_stats.asmd(
                pd.DataFrame({"a": a1, "b[A]": b1_A, "b[B]": b1_B}),
                pd.DataFrame({"a": a2, "b[A]": b2_A, "b[B]": b2_B}),
                w1,
                w2,
                "target",
                True,
            ).to_dict()

            print(r)
            # {'a': 2.82842712474619, 'b': 0.7071067811865475, 'mean(asmd)': 1.7677669529663689}

    """
    if not isinstance(sample_df, pd.DataFrame):
        raise ValueError(f"sample_df must be pd.DataFrame, is {type(sample_df)}")
    if not isinstance(target_df, pd.DataFrame):
        raise ValueError(f"target_df must be pd.DataFrame, is {type(target_df)}")
    possible_std_type = ("sample", "target", "pooled")
    if not (std_type in possible_std_type):
        raise ValueError(f"std_type must be in {possible_std_type}, is {std_type}")
    if sample_df.columns.values.tolist() != target_df.columns.values.tolist():
        logger.warning(
            f"""
            sample_df and target_df must have the same column names.
            sample_df column names: {sample_df.columns.values.tolist()}
            target_df column names: {target_df.columns.values.tolist()}"""
        )

    sample_mean = descriptive_stats(sample_df, sample_weights, "mean")
    target_mean = descriptive_stats(target_df, target_weights, "mean")

    if std_type == "sample":
        std = descriptive_stats(sample_df, sample_weights, "std")
    elif std_type == "target":
        std = descriptive_stats(target_df, target_weights, "std")
    elif std_type == "pooled":
        target_std = descriptive_stats(target_df, target_weights, "std")
        sample_std = descriptive_stats(sample_df, sample_weights, "std")
        std = np.sqrt(((sample_std**2) + (target_std**2)) / 2)

    out = abs(sample_mean - target_mean) / std  # pyre-ignore[61]: std is always defined

    #  Remove na indicator columns; it's OK to assume that these columns are
    #  indicators because add_na_indicator enforces it
    out = out.loc[:, (c for c in out.columns.values if not c.startswith("_is_na_"))]
    out = out.replace([np.inf, -np.inf], np.nan)

    # TODO (p2): verify that df column names are unique (otherwise throw an exception).
    #            it should probably be upstream during in the Sample creation process.
    weights = _weights_per_covars_names(out.columns.values.tolist())[
        ["weight"]
    ].transpose()
    out = pd.concat((out, weights))
    mean = weighted_mean(out.iloc[0], out.loc["weight"])
    out["mean(asmd)"] = mean

    out = out.iloc[0]
    out.name = None

    if aggregate_by_main_covar:
        out = _aggregate_asmd_by_main_covar(out)

    return out


def _aggregate_asmd_by_main_covar(asmd_series: pd.Series) -> pd.Series:
    """
    This function helps to aggregate the ASMD, which is broken down per level for a category variable, into one value per covariate.
    This is useful since it allows us to get high level view of features such as country, locale, etc.
    It also allows us to filter out variables (such as is_today) before the overall averaging of the ASMD.

    Args:
        asmd_series (pd.Series): a value for each covariate. Covariate name are in the index, the values are the asmd values.

    Returns:
        pd.Series: If asmd_series had several items broken by one-hot encoding,
            then they would be averaged (with equal weight to each).

    Examples:
        ::

            from balance.stats_and_plots.weighted_comparisons_stats import _aggregate_asmd_by_main_covar

            asmd_series = pd.Series(
            {
                'age': 0.5,
                'education[T.high_school]': 1,
                'education[T.bachelor]': 2,
                'education[T.masters]': 3,
                'education[T.phd]': 4,
            })

            _aggregate_asmd_by_main_covar(asmd_series).to_dict()

            # output:
            # {'age': 0.5, 'education': 2.5}
    """
    weights = _weights_per_covars_names(asmd_series.index.values.tolist())

    # turn things into DataFrame to make it easy to aggregate.
    out = pd.concat((asmd_series, weights), axis=1)

    def _weighted_mean_for_our_df(x: pd.DataFrame) -> pd.Series:
        values = x.iloc[:, 0]
        weights = x["weight"]
        weighted_mean = pd.Series(
            ((values * weights) / weights.sum()).sum(), index=["mean"]
        )
        return weighted_mean

    out = out.groupby("main_covar_names").apply(_weighted_mean_for_our_df).iloc[:, 0]
    out.name = None
    out.index.name = None

    return out


# TODO: (p2) sample_before and sample_after are redundant, the moment weights of
#           before and after are supplied directly.
#           In the future, we can either omit sample_after, or change the names to
#           reflect a support a comparison of two panels to some target populations.
def asmd_improvement(
    sample_before: pd.DataFrame,
    sample_after: pd.DataFrame,
    target: pd.DataFrame,
    sample_before_weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    sample_after_weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    target_weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
) -> np.float64:
    """Calculates the improvement in mean(asmd) from before to after applying some weight adjustment.

    Args:
        sample_before (pd.DataFrame): DataFrame of the sample before adjustments.
        sample_after (pd.DataFrame): DataFrame of the sample after adjustments
            (should be identical to sample_before. But could be used to compare two populations).
        target (pd.DataFrame): DataFrame of the target population.
        sample_before_weights (Union[ List, pd.Series, np.ndarray, ], optional): Weights before adjustments (i.e.: design weights). Defaults to None.
        sample_after_weights (Union[ List, pd.Series, np.ndarray, ], optional): Weights after some adjustment. Defaults to None.
        target_weights (Union[ List, pd.Series, np.ndarray, ], optional): Design weights of the target population. Defaults to None.

    Returns:
        np.float64: The improvement is taking the (before_mean_asmd-after_mean_asmd)/before_mean_asmd.
        The asmd is calculated using :func:`asmd`.
    """
    asmd_mean_before = asmd(
        sample_before, target, sample_before_weights, target_weights
    ).loc["mean(asmd)"]
    asmd_mean_after = asmd(
        sample_after, target, sample_after_weights, target_weights
    ).loc["mean(asmd)"]
    return (asmd_mean_before - asmd_mean_after) / asmd_mean_before


def outcome_variance_ratio(
    df_numerator: pd.DataFrame,
    df_denominator: pd.DataFrame,
    w_numerator: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    w_denominator: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
) -> pd.Series:
    """Calculate ratio of weighted variances of two DataFrames

    Directly calculating the empirical ratio of variance of the outcomes before and after weighting.
    Notice that this is different than design effect.
    The Deff estimates the ratio of variances of the weighted means, while this function calculates the ratio
    of empirical weighted variance of the data.

    Args:
        df_numerator (pd.DataFrame): df_numerator
        df_denominator (pd.DataFrame): df_denominator
        w_numerator (Union[ List, pd.Series, np.ndarray, None, ], optional): w_numerator. Defaults to None.
        w_denominator (Union[ List, pd.Series, np.ndarray, None, ], optional): w_denominator. Defaults to None.

    Returns:
        pd.Series: (np.float64) A series of calculated ratio of variances for each outcome.
    """
    numerator_w_var = weighted_var(df_numerator, w_numerator)
    denominator_w_var = weighted_var(df_denominator, w_denominator)
    return numerator_w_var / denominator_w_var
